import numbers
import warnings
from collections import namedtuple

import dask
import dask.array as da
import dask.dataframe as dd
import geopandas as gpd
import numpy as np
import xarray as xr
from affine import Affine
from numba import jit
from shapely.geometry import Point

from raster_tools.dask_utils import dask_nanmax, dask_nanmin
from raster_tools.dtypes import (
    DTYPE_INPUT_TO_DTYPE,
    is_bool,
    is_float,
    is_int,
    is_scalar,
    is_str,
    promote_dtype_to_float,
    should_promote_to_fit,
)
from raster_tools.masking import (
    create_null_mask,
    get_default_null_value,
    reconcile_nullvalue_with_dtype,
)
from raster_tools.utils import can_broadcast, make_raster_ds, merge_masks

from .io import (
    IO_UNDERSTOOD_TYPES,
    RasterDataError,
    RasterIOError,
    chunk,
    is_batch_file,
    normalize_xarray_data,
    open_raster_from_path,
    write_raster,
)


class RasterDeviceMismatchError(BaseException):
    pass


class RasterDeviceError(BaseException):
    pass


class RasterNoDataError(BaseException):
    pass


_REDUCTION_FUNCS = (
    "all",
    "any",
)
_NAN_REDUCTION_FUNCS = (
    "max",
    "mean",
    "min",
    "prod",
    "std",
    "sum",
    "var",
)


_REDUCTION_DOCSTRING = """\
    Reduce the raster to a single dask value by applying `{}` across all bands.

    All arguments are ignored.
"""


def _inject_reductions(cls):
    pass
    funcs = [(name, getattr(np, name)) for name in _REDUCTION_FUNCS]
    funcs += [
        (name, getattr(np, "nan" + name)) for name in _NAN_REDUCTION_FUNCS
    ]
    for name, f in funcs:
        func = cls._build_reduce_method(f)
        func.__name__ = name
        func.__doc__ = _REDUCTION_DOCSTRING.format(name)
        setattr(cls, name, func)


_MIN_MAX_FUNC_MAP = {
    np.max: dask_nanmax,
    np.nanmax: dask_nanmax,
    np.min: dask_nanmin,
    np.nanmin: dask_nanmin,
}


class _ReductionsMixin:
    """
    This mixin class adds reduction methods like `all`, `sum`, etc to the
    Raster class. Having these methods also allows numpy reduction functions to
    use them for dynamic dispatch. So `np.sum(raster)` will call the class'
    `sum` method. All reductions return dask results.
    """

    __slots__ = ()

    def __init_subclass__(cls, **kwargs):
        _inject_reductions(cls)
        super().__init_subclass__(**kwargs)

    @classmethod
    def _build_reduce_method(cls, func):
        if func in _MIN_MAX_FUNC_MAP:
            func = _MIN_MAX_FUNC_MAP[func]

        def method(self, *args, **kwargs):
            # Take args and kwargs to stay compatible with numpy
            data = self.data
            values = data[~self._ds.mask.data]
            return func(values)

        return method


def _normalize_ufunc_other(other, this):
    if other is None or isinstance(other, numbers.Number):
        return other

    if isinstance(other, xr.Dataset):
        return Raster(other)
    elif isinstance(other, (list, tuple, range, np.ndarray, da.Array)):
        # allow only if broadcastable
        if isinstance(other, da.Array) and np.isnan(other.size):
            raise ValueError("Dask arrays must have known chunks")
        other = np.atleast_1d(other)
        if not can_broadcast(this.shape, other.shape):
            msg = (
                "Raster could not broadcast together with array of shape"
                f" {other.shape}."
            )
            if other.squeeze().ndim == 1 and other.size == this.shape[0]:
                msg += " Did you mean to use Raster.bandwise?"
            raise ValueError(msg)
        return other
    elif isinstance(other, xr.DataArray):
        return Raster(other)
    else:
        assert isinstance(other, Raster)
    return other


def _apply_ufunc(ufunc, this, left, right=None, kwargs=None, out=None):
    args = [left]
    if right is not None:
        args.append(right)
    other = left if left is not this else right
    types = [getattr(a, "dtype", type(a)) for a in args]
    masked = any(getattr(r, "_masked", False) for r in args)
    ufunc_args = [getattr(a, "xdata", a) for a in args]
    kwargs = kwargs or {}
    out_crs = None
    if this.crs is not None:
        out_crs = this.crs
    elif isinstance(other, Raster) and other.crs is not None:
        out_crs = other.crs

    ufname = ufunc.__name__
    if ufname.startswith("bitwise") and any(is_float(t) for t in types):
        raise TypeError(
            "Bitwise operations are not compatible with float dtypes. You may"
            " want to convert one or more inputs to a boolean or integer dtype"
            " (e.g. 'raster > 0' or 'raster.astype(int)')."
        )
    if ufname.endswith("shift") and any(is_float(t) for t in types):
        raise TypeError(
            "right_shift and left_shift operations are not compatible with"
            " float dtypes. You may want to convert one or more inputs to an"
            " integer dtype (e.g. 'raster.astype(int)')."
        )

    xr_out = ufunc(*ufunc_args, **kwargs)
    multi_out = ufunc.nout > 1
    if not masked:
        mask = create_null_mask(xr_out, None)
        if not multi_out:
            xmask = xr.DataArray(mask, dims=xr_out.dims, coords=xr_out.coords)
            xr_out = xr_out.rio.write_nodata(None)
            ds_out = make_raster_ds(xr_out, xmask)
        else:
            xmask = xr.DataArray(
                mask, dims=xr_out[0].dims, coords=xr_out[0].coords
            )
            xr_out = [x.rio.write_nodata(None) for x in xr_out]
            ds_out = [make_raster_ds(x, xmask) for x in xr_out]
    else:
        xmask = merge_masks([r.xmask for r in args if isinstance(r, Raster)])
        if not multi_out:
            nv = get_default_null_value(xr_out.dtype)
            xr_out = xr.where(xmask, nv, xr_out).rio.write_nodata(nv)
            ds_out = make_raster_ds(xr_out, xmask)
        else:
            nvs = [get_default_null_value(x.dtype) for x in xr_out]
            xr_out = [
                xr.where(xmask, nv, x).rio.write_nodata(nv)
                for x, nv in zip(xr_out, nvs)
            ]
            ds_out = [make_raster_ds(x, xmask) for x in xr_out]
    if out_crs is not None:
        if multi_out:
            ds_out = [x.rio.write_crs(out_crs) for x in ds_out]
        else:
            ds_out = ds_out.rio.write_crs(out_crs)

    if out is not None:
        # "Inplace"
        out._ds = ds_out
        return out

    if not multi_out:
        return Raster(ds_out, _fast_path=True)

    rs_outs = tuple(Raster(x, _fast_path=True) for x in ds_out)
    return rs_outs


_UNARY_UFUNCS = frozenset(
    (np.absolute, np.invert, np.logical_not, np.negative, np.positive)
)
_UNSUPPORED_UFUNCS = frozenset((np.isnat, np.matmul))


class _RasterBase(np.lib.mixins.NDArrayOperatorsMixin, _ReductionsMixin):
    """This class implements methods for handling numpy ufuncs."""

    __slots__ = ()

    _HANDLED_TYPES = (
        numbers.Number,
        xr.DataArray,
        np.ndarray,
        da.Array,
        list,
        tuple,
        range,
    )
    # Higher than xarray objects
    __array_priority__ = 70

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        out = kwargs.pop("out", ())
        for x in inputs + out:
            if not isinstance(x, self._HANDLED_TYPES + (_RasterBase,)):
                return NotImplemented

        if ufunc in _UNSUPPORED_UFUNCS:
            raise TypeError(
                f"Raster objects are not supported for ufunc: '{ufunc}'."
            )

        if ufunc.signature is not None:
            raise NotImplementedError("Raster does not support gufuncs")

        if method != "__call__":
            raise NotImplementedError(
                f"{method} for ufunc {ufunc} is not implemented on Raster "
                "objects."
            )

        if len(inputs) > ufunc.nin:
            raise TypeError(
                "Too many inputs for ufunc:"
                f" inputs={len(inputs)}, ufunc={ufunc!r}"
            )

        if len(out):
            if len(out) > 1:
                raise NotImplementedError(
                    "The 'out' keyword is only supported for inplace "
                    "operations, not operations with multiple outputs."
                )
            (out,) = out
            if out is not self:
                raise NotImplementedError(
                    "'out' must be the raster being operated on."
                )
        else:
            out = None

        left = inputs[0]
        right = inputs[1] if ufunc.nin == 2 else None
        other = left if left is not self else right
        other = _normalize_ufunc_other(other, self)
        if self is left:
            right = other
        else:
            left = other
        return _apply_ufunc(ufunc, self, left, right=right, out=out)

    def __array__(self, dtype=None):
        return self._ds.raster.__array__(dtype)


def _normalize_bandwise_other(other, target_shape):
    other = np.atleast_1d(other)
    if isinstance(other, da.Array) and np.isnan(other.size):
        raise ValueError("Dask arrays must have known chunks")
    if other.size == 1:
        return other.ravel()
    if can_broadcast(other.shape, target_shape):
        raise ValueError(
            "Operands can be broadcast together. A bandwise operation is not"
            " needed here."
        )
    if other.ndim == 1 and other.size == target_shape[0]:
        # Map a list/array of values that has same length as # of bands to
        # each band
        return other.reshape((-1, 1, 1))
    raise ValueError(
        "Received array with incompatible shape in bandwise operation: "
        f"{other.shape}"
    )


class BandwiseOperationAdapter(np.lib.mixins.NDArrayOperatorsMixin):
    __slots__ = ("_raster",)

    _HANDLED_TYPES = (
        numbers.Number,
        np.ndarray,
        da.Array,
        list,
        tuple,
        range,
    )
    # Higher than xarray objects
    __array_priority__ = 70

    def __init__(self, raster):
        self._raster = raster

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        out = kwargs.pop("out", None)
        for x in inputs:
            if not isinstance(
                x,
                self._HANDLED_TYPES
                + (
                    Raster,
                    BandwiseOperationAdapter,
                ),
            ):
                return NotImplemented

        if len(inputs) < 2:
            raise ValueError(
                "Bandwise operations are binary and require 2 inputs. Only"
                " one given."
            )
        if len(inputs) > 2:
            raise ValueError(
                "Bandwise operations are binary and require 2 inputs."
                f" {len(inputs)} given."
            )

        if ufunc in _UNSUPPORED_UFUNCS:
            raise TypeError(
                "The given ufunc is not supported for bandwise operations:"
                f" {ufunc!r}."
            )

        if ufunc.signature is not None:
            raise NotImplementedError("Raster does not support gufuncs")

        if method != "__call__":
            raise NotImplementedError(
                f"{method} for ufunc {ufunc} is not implemented for bandwise "
                " operations."
            )

        if len(inputs) > ufunc.nin:
            raise TypeError(
                "Too many inputs for ufunc:"
                f" inputs={len(inputs)}, ufunc={ufunc!r}"
            )

        if out is not None:
            raise NotImplementedError(
                "The 'out' keyword is not supported for bandwise operations"
            )

        left, right = inputs
        other = left if left is not self else right
        other = _normalize_bandwise_other(other, self._raster.shape)
        if self is left:
            left = self._raster
            right = other
        else:
            left = other
            right = self._raster

        return ufunc(left, right, **kwargs)


def _array_to_xarray(ar):
    if len(ar.shape) > 3 or len(ar.shape) < 2:
        raise ValueError(f"Invalid raster shape for numpy array: {ar.shape}")
    if len(ar.shape) == 2:
        # Add band dim
        ar = ar[None]
    if not isinstance(ar, da.Array):
        has_nan = is_float(ar.dtype) and np.isnan(ar).any()
    else:
        has_nan = is_float(ar.dtype)
    # The band dimension needs to start at 1 to match raster conventions
    coords = [np.arange(1, ar.shape[0] + 1)]
    # Add 0.5 offset to move coords to center of cell
    # y
    coords.append(np.arange(ar.shape[1])[::-1] + 0.5)
    coords.append(np.arange(ar.shape[2]) + 0.5)
    xrs = xr.DataArray(ar, dims=["band", "y", "x"], coords=coords)
    if has_nan:
        xrs.attrs["_FillValue"] = np.nan
    else:
        # No way to know null value
        xrs.attrs["_FillValue"] = None
    return xrs


def _try_to_get_null_value_xarray(xrs):
    if (
        "_FillValue" in xrs.attrs
        and xrs.attrs["_FillValue"] is None
        and xrs.rio.nodata is None
        and xrs.rio.encoded_nodata is None
        and not dask.is_dask_collection(xrs)
    ):
        return None
    null = xrs.attrs.get("_FillValue", None)
    if null is not None:
        return null
    nv1 = xrs.rio.nodata
    nv2 = xrs.rio.encoded_nodata
    # All are None, return nan for float and None otherwise
    if all(nv is None for nv in [nv1, nv2]):
        if is_float(xrs.dtype):
            if dask.is_dask_collection(xrs) or np.isnan(xrs.data).any():
                return np.nan
            else:
                return None
        else:
            return None
    # All are not None, return nv1
    if all(nv is not None for nv in [nv1, nv2]):
        return nv1
    # One is None and the other is not, return the valid null value
    return nv1 if nv1 is not None else nv2


def _dataarry_to_raster_ds(xin):
    nv = _try_to_get_null_value_xarray(xin)
    xin = normalize_xarray_data(xin)
    if not dask.is_dask_collection(xin) or xin.data.chunksize[0] != 1:
        xin = chunk(xin)

    if nv is None:
        mask = create_null_mask(xin, nv)
        mask = xr.DataArray(
            create_null_mask(xin, nv), dims=xin.dims, coords=xin.coords
        )
        if xin.rio.crs is not None:
            mask = mask.rio.write_crs(xin.rio.crs)
    elif np.isnan(nv):
        mask = np.isnan(xin)
    else:
        mask = xin == nv

    orig_nv = nv
    nv = reconcile_nullvalue_with_dtype(nv, xin.dtype)
    crs = xin.rio.crs
    if nv is not None and (np.isnan(orig_nv) or orig_nv != nv):
        xin = xr.where(mask, nv, xin)
        if crs is not None:
            xin = xin.rio.write_crs(crs)
    xin = xin.rio.write_nodata(nv)
    return make_raster_ds(xin, mask)


def _dataset_to_raster_ds(xin):
    if (
        dask.is_dask_collection(xin)
        and set(xin.data_vars) == set(["raster", "mask"])
        and dask.is_dask_collection(xin.raster)
        and dask.is_dask_collection(xin.mask)
        and is_bool(xin.mask.dtype)
        and xin.raster.dims == ("band", "y", "x")
        and xin.mask.dims == ("band", "y", "x")
        and xin.raster.chunks == xin.mask.chunks
        and xin.raster.rio.crs == xin.mask.rio.crs
    ):
        return xin

    nvars = len(xin.data_vars)
    if nvars < 2 or nvars > 2 or set(xin.data_vars) != set(["raster", "mask"]):
        raise ValueError(
            "Dataset input must have only 'raster' and 'mask' variables"
        )

    raster = xin.raster
    mask = xin.mask
    if not is_bool(mask.dtype):
        raise TypeError("mask must have boolean dtype")

    raster = normalize_xarray_data(raster)
    if not dask.is_dask_collection(raster):
        raster = chunk(raster)
    mask = normalize_xarray_data(mask)
    if mask.shape != raster.shape:
        raise ValueError("raster and mask dimensions do not match")
    if not dask.is_dask_collection(mask):
        mask = mask.chunk(chunks=raster.chunks)

    nv = _try_to_get_null_value_xarray(raster)
    if nv is None:
        nv = get_default_null_value(raster.dtype)
    else:
        nv = reconcile_nullvalue_with_dtype(nv, raster.dtype)
    if nv is not None:
        raster.data = xr.where(mask, nv, raster.data)
    raster.rio.write_nodata(nv, inplace=True)
    crs = raster.rio.crs
    if crs != mask.rio.crs:
        mask.rio.write_crs(crs, inplace=True)
    return make_raster_ds(raster, mask)


def _xarray_to_raster_ds(xin):
    if isinstance(xin, xr.DataArray):
        return _dataarry_to_raster_ds(xin)
    return _dataset_to_raster_ds(xin)


def get_raster_ds(raster):
    # Copy to take ownership
    if isinstance(raster, Raster):
        ds = raster._ds.copy(deep=True)
    elif isinstance(raster, (xr.DataArray, xr.Dataset)):
        ds = _xarray_to_raster_ds(raster.copy())
    elif isinstance(raster, (np.ndarray, da.Array)):
        ds = _xarray_to_raster_ds(_array_to_xarray(raster.copy()))
    elif type(raster) in IO_UNDERSTOOD_TYPES:
        if is_batch_file(raster):
            # Import here to avoid circular import errors
            from raster_tools.batch import parse_batch_script

            ds = parse_batch_script(raster).final_raster._ds
        else:
            rs, mask, nv = open_raster_from_path(raster)
            xmask = xr.DataArray(mask, dims=rs.dims, coords=rs.coords)
            ds = make_raster_ds(rs.rio.write_nodata(nv), xmask)
            ds = _xarray_to_raster_ds(ds)
    else:
        raise TypeError(f"Could not resolve input to a raster: {raster!r}")
    return ds


RasterQuadrantsResult = namedtuple(
    "RasterQuadrantsResult", ("nw", "ne", "sw", "se")
)


class Raster(_RasterBase):
    """Abstraction of georeferenced raster data with lazy function evaluation.

    Raster is a wrapper around xarray Datasets and DataArrays. It takes
    advantage of xarray's dask integration to allow lazy loading and
    evaluation. It allows a pipeline of operations on underlying raster
    sources to be built in a lazy fashion and then evaluated effiently.
    Most mathematical operations have been overloaded so operations such as
    ``z = x - y`` and ``r = x**2`` are possible.

    All operations on a Raster return a new Raster.

    Parameters
    ----------
    raster : str, Raster, xarray.DataArray, xarray.Dataset, numpy.ndarray
        The raster source to use for this Raster. If `raster` is a string,
        it is treated like a path. If `raster` is a Raster, a copy is made
        and its raster source is used. If `raster` is an xarray DataArray,
        Dataset, or numpy array, it is used as the source. Dataset objects must
        contain 'raster' and 'mask' variables. The dimensions of these
        variables are assumed to be "band", "y", "x", in that order. The 'mask'
        variable must have boolean dtype.

    """

    __slots__ = ("_ds",)

    def __init__(self, raster, _fast_path=False):
        if _fast_path:
            self._ds = raster
        else:
            self._ds = get_raster_ds(raster)

    def __repr__(self):
        # TODO: implement
        return repr(self._ds.raster)

    @property
    def null_value(self):
        """The raster's null value used to fill missing or invalid entries."""
        return self._ds.raster.rio.nodata

    @property
    def _masked(self):
        return self.null_value is not None

    @property
    def xdata(self):
        """The underlying :class:`xarray.DataArray` (read only)"""
        return self._ds.raster

    @property
    def data(self):
        """The underlying dask array of data"""
        return self._ds.raster.data

    @property
    def values(self):
        """
        The raw internal raster as a numpy array.

        .. note::
           This triggers computation and loads the raster data into memory.
        """
        return self._ds.raster.values

    @property
    def mask(self):
        """The null value mask as a dask array."""
        return self._ds.mask.data

    @property
    def xmask(self):
        """The null value mask as an xarray DataArray."""
        return self._ds.mask

    @property
    def band(self):
        """The band coordinate values."""
        return self._ds.band.data

    @property
    def x(self):
        """The x (horizontal) coordinate values."""
        return self._ds.x.data

    @property
    def y(self):
        """The y (vertical) coordinate values."""
        return self._ds.y.data

    @property
    def dtype(self):
        """The dtype of the data."""
        return self._ds.raster.dtype

    @property
    def shape(self):
        """The shape of the underlying raster. Dim 0 is the band dimension.

        The shape will always be of the form ``(B, Y, X)`` where ``B`` is the
        band dimension, ``Y`` is the y dimension, and ``X`` is the x dimension.

        """
        return self._ds.raster.shape

    @property
    def nbands(self):
        """The number of bands."""
        return self._ds.band.size

    @property
    def crs(self):
        """The raster's CRS.

        This is a :obj:`rasterio.crs.CRS` object."""
        return self._ds.rio.crs

    @property
    def affine(self):
        """The affine transformation for the raster data.

        This is an :obj:`affine.Affine` object.

        """
        return self._ds.rio.transform(True)

    @property
    def resolution(self):
        """The x and y cell sizes as a tuple. Values are always positive."""
        return self._ds.rio.resolution(True)

    @property
    def bounds(self):
        """Bounds tuple (minx, miny, maxx, maxy)"""
        r, c = self.shape[1:]
        minx, maxy = self.xy(0, 0, offset="ul")
        maxx, miny = self.xy(r - 1, c - 1, "lr")
        return (minx, miny, maxx, maxy)

    @property
    def bandwise(self):
        """Returns an adapter for band-wise operations on this raster

        The returned adapter allows lists/arrays of values to be mapped to the
        raster's bands for binary operations.

        """
        return BandwiseOperationAdapter(self)

    @property
    def _rs(self):
        warnings.warn(
            "'Raster._rs' is deprecated. It will soon be removed. Use xdata.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._ds.raster

    @property
    def _data(self):
        warnings.warn(
            "'Raster._data' is deprecated. It will soon be removed. Use data.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._ds.raster.data

    @property
    def _values(self):
        warnings.warn(
            "'Raster._values' is deprecated. It will soon be removed. Use "
            "values.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._ds.raster.values

    @property
    def _mask(self):
        warnings.warn(
            "'Raster._mask' is deprecated. It will soon be removed. Use mask.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._ds.mask.data

    @property
    def _null_value(self):
        warnings.warn(
            "'Raster._null_value' is deprecated. It will soon be removed. Use "
            "null_value.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.null_value

    @property
    def xrs(self):
        warnings.warn(
            "'Raster.xrs' is deprecated. It will soon be removed. Use xdata.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._ds.raster

    @property
    def pxrs(self):
        warnings.warn(
            "'Raster.pxrs' is deprecated. It will soon be removed.",
            DeprecationWarning,
            stacklevel=2,
        )
        return get_raster(self, null_to_nan=True)._ds.raster

    def chunk(self, chunks):
        """Rechunk the underlying raster

        Parameters
        ----------
        chunks : tuple of int
            The chunk sizes for each dimension ("band", "y", "x", in that
            order). Must be a 3-tuple.

        Returns
        -------
        Raster
            The rechunked raster

        """
        assert len(chunks) == 3
        return Raster(
            self._ds.chunk(
                {d: cs for d, cs in zip(["band", "y", "x"], chunks)}
            ),
            _fast_path=True,
        )

    def to_dataset(self):
        """Returns the underlying `xarray.Dataset`."""
        return self._ds

    def plot(self, *args, **kwargs):
        """
        Plot the raster. Args and kwargs are passed to xarray's plot function.
        """
        return get_raster(self, null_to_nan=True).xdata.plot(*args, **kwargs)

    def get_chunked_coords(self):
        """Get lazy coordinate arrays, in x-y order, chunked to match data."""
        xc = da.from_array(self.x, chunks=self.data.chunks[2]).reshape(
            (1, 1, -1)
        )
        yc = da.from_array(self.y, chunks=self.data.chunks[1]).reshape(
            (1, -1, 1)
        )
        return xc, yc

    def close(self):
        """Close the underlying source"""
        self._ds.close()

    def save(self, path, no_data_value=None, **gdal_kwargs):
        """Compute the final raster and save it to the provided location.

        Parameters
        ----------
        path : str
            The target location to save the raster to.
        no_data_value : scalar, optional
            A new null value to use when saving.
        **gdal_kwargs : kwargs, optional
            Additional keyword arguments to to pass to rasterio and GDAL when
            writing the raster data.

        Returns
        -------
        Raster
            A raster pointing to the saved location.

        """
        # TODO: warn of overwrite
        rs = self
        if no_data_value is not None:
            rs = self.set_null_value(no_data_value)
        if rs._masked and is_bool(rs.dtype):
            # Cast to uint and burn in mask
            rs = rs.astype("uint8").set_null_value(
                get_default_null_value("uint8")
            )
        xrs = rs.xdata
        write_raster(
            xrs,
            path,
            rs.null_value if no_data_value is None else no_data_value,
            **gdal_kwargs,
        )
        return Raster(path)

    def eval(self):
        """Compute any applied operations and return the result as new Raster.

        Note that the unerlying sources will be loaded into memory for the
        computations and the result will be fixed in memory. The original
        Raster will be unaltered.

        """
        ds = chunk(self._ds.compute())
        # A new raster is returned to mirror the xarray and dask APIs
        return Raster(ds)

    def copy(self):
        """Returns a copy of this Raster."""
        return Raster(self)

    def astype(self, dtype, warn_about_null_change=True):
        """Return a copy of the Raster cast to the specified type.

        Parameters
        ----------
        dtype : str, type, numpy.dtype
            The new type to cast the raster to. The null value will also be
            cast to this type. If the null value cannot be safely cast to the
            new type, a default null value for the given `dtype` is used. This
            will produce a warning unless `warn_about_null_change` is set to
            ``False``.
        warn_about_null_change : bool, optional
            Can be used to silence warnings. The default is to always warn
            about null value changes.

        Returns
        -------
        Raster
            The new `dtype` raster.

        """
        if dtype not in DTYPE_INPUT_TO_DTYPE:
            raise ValueError(f"Unsupported type: '{dtype}'")
        dtype = DTYPE_INPUT_TO_DTYPE[dtype]

        if dtype == self.dtype:
            return self.copy()

        xrs = self._ds.raster
        nv = self.null_value
        mask = self._ds.mask

        xrs = xrs.astype(dtype)
        if self._masked:
            nv = reconcile_nullvalue_with_dtype(
                nv, dtype, warn_about_null_change
            )
            if nv != self.null_value:
                xrs = xr.where(mask, nv, xrs).rio.write_nodata(nv)
        ds = make_raster_ds(xrs, mask)
        if self.crs is not None:
            ds = ds.rio.write_crs(self.crs)
        return Raster(ds, _fast_path=True)

    def get_bands(self, bands):
        """Retrieve the specified bands as a new Raster. Indexing starts at 1.

        Parameters
        ----------
        bands : int or sequence of ints
            The band or bands to be retrieved. A single band Raster is returned
            if `bands` is an int and a multiband Raster is returned if it is a
            sequence. The band numbers may be out of order and contain repeated
            elements. if `bands` is a sequence, the multiband raster's bands
            will be in the order provided.

        Returns
        -------
        Raster
            The resulting raster composed of the specified bands in the order
            given.

        """
        if is_int(bands):
            bands = [bands]
        if not all(is_int(b) for b in bands):
            raise TypeError("All band numbers must be integers")
        bands = list(bands)
        if len(bands) == 0:
            raise ValueError("No bands provided")
        if any(b < 1 or b > self.nbands for b in bands):
            raise IndexError(
                f"One or more band numbers were out of bounds: {bands}"
            )

        if len(bands) == 1 and self.nbands == 1:
            return self.copy()
        # We have to do sel() followed by concat() in order to get size 1
        # chunking along the band dim
        dss = [self._ds.sel(band=b) for b in bands]
        ds = xr.concat(dss, dim="band")
        # Reset band numbering
        ds["band"] = np.arange(len(bands)) + 1
        return Raster(ds, _fast_path=True)

    def set_crs(self, crs):
        """Set the CRS for the underlying data.

        Parameters
        ----------
        crs : object
            The desired CRS. This can be anything accepted by
            :obj:`rasterio.crs.CRS.from_user_input` (i.e. `4326`,
            `"epsg:4326"`, etc).

        Returns
        -------
        Raster
            A new raster with the specified CRS.

        """
        return Raster(self._ds.rio.write_crs(crs), _fast_path=True)

    def set_null_value(self, value):
        """Sets or replaces the null value for the raster.

        If there was previously no null value for the raster, one is set. If
        there was already a null value, then it is replaced. If the raster has
        an integer dtype and `value` is a float, the dtype will be promoted to
        a float dtype. If `value` is None, the null value is cleared. The
        raster data is not changed in this case.

        Parameters
        ----------
        value : scalar or None
            The new null value for the raster. If None, the resulting raster
            will have no null value, i.e. the null value is cleared.

        Returns
        -------
        Raster
            The new resulting Raster.

        """
        if value is not None and not (is_scalar(value) or is_bool(value)):
            raise TypeError(f"Value must be a scalar or None: {value}")

        xrs = self._ds.raster.copy()
        # Cast up to float if needed
        if should_promote_to_fit(self.dtype, value):
            xrs = xrs.astype(promote_dtype_to_float(self.dtype))

        if value is None:
            # Burn in current mask values and then clear null value
            # TODO: burning should not be needed as the values should already
            # be set
            xrs = self.burn_mask().xdata.rio.write_nodata(None)
            mask = xr.zeros_like(xrs, dtype=bool)
            return Raster(make_raster_ds(xrs, mask), _fast_path=True)

        # Update mask
        mask = self._ds.mask
        temp_mask = np.isnan(xrs) if np.isnan(value) else xrs == value
        if self._masked:
            mask = mask | temp_mask
        else:
            mask = temp_mask
        xrs = xrs.rio.write_nodata(value)
        return Raster(make_raster_ds(xrs, mask), _fast_path=True).burn_mask()

    def burn_mask(self):
        """Fill null-masked cells with null value.

        Use this as a way to return the underlying data to a known state. If
        dtype is boolean, the null cells are set to ``True`` instead of
        promoting to fit the null value.
        """
        if not self._masked:
            return self
        nv = self.null_value
        if is_bool(self.dtype):
            nv = get_default_null_value(self.dtype)
        # call write_nodata because xr.where drops the nodata info
        xrs = xr.where(self._ds.mask, nv, self._ds.raster).rio.write_nodata(nv)
        if self.crs is not None:
            xrs = xrs.rio.write_crs(self.crs)
        return Raster(make_raster_ds(xrs, self._ds.mask), _fast_path=True)

    def to_null_mask(self):
        """
        Returns a boolean Raster with True at null values and False otherwise.

        Returns
        -------
        Raster
            The resulting mask Raster. It is True where this raster contains
            null values and False everywhere else.

        """
        return Raster(
            make_raster_ds(
                self._ds.mask.rio.write_nodata(None),
                xr.zeros_like(self._ds.mask),
            ),
            _fast_path=True,
        )

    def replace_null(self, value):
        """Replaces null values with `value`.

        Parameters
        ----------
        value : scalar
            The new value to replace null values with.

        Returns
        -------
        Raster
            The new resulting Raster. If `value` is a float and the raster
            dtype is int, the raster type will be promoted to float.

        """
        if not is_scalar(value):
            raise TypeError("value must be a scalar")

        return self.set_null_value(value)

    def where(self, condition, other):
        """Filter elements from this raster according to `condition`.

        Parameters
        ----------
        condition : str or Raster
            A boolean or int raster that indicates where elements in this
            raster should be preserved and where `other` should be used. If
            the condition is an int raster, it is coerced to bool using
            `condition > 0`.  ``True`` cells pull values from this raster and
            ``False`` cells pull from `other`. *str* is treated as a path to a
            raster.
        other : scalar, str or Raster
            A raster or value to use in locations where `condition` is
            ``False``. *str* is treated as a path to a raster.

        Returns
        -------
        Raster
            The resulting filtered Raster.

        """
        from raster_tools.general import where

        return where(condition, self, other)

    def remap_range(self, mapping, inclusivity="left"):
        """Remaps values based on a mapping or list of mappings.

        Mappings are applied all at once with earlier mappings taking
        precedence.

        Parameters
        ----------
        mapping : 3-tuple of scalars or list of 3-tuples of scalars
            A tuple or list of tuples containing ``(min, max, new_value)``
            scalars. The mappiing(s) map values between the min and max to the
            ``new_value``. If `mapping` is a list and there are mappings that
            conflict or overlap, earlier mappings take precedence.
            `inclusivity` determines which sides of the range are inclusive and
            exclusive.
        inclusivity : str, optional
            Determines whether to be inclusive or exclusive on either end of
            the range. Default is `'left'`.

            'left' [min, max)
                Left (min) side is inclusive and right (max) side is exclusive.
            'right' (min, max]
                Left (min) side is exclusive and right (max) side is inclusive.
            'both' [min, max]
                Both sides are inclusive.
            'none' (min, max)
                Both sides are exclusive.

        Returns
        -------
        Raster
            The resulting Raster.

        See Also
        --------
        raster_tools.general.remap_range

        """
        # local import to avoid circular import
        from raster_tools.general import remap_range

        return remap_range(self, mapping, inclusivity=inclusivity)

    def reclassify(self, remapping, unmapped_to_null=False):
        """Reclassify raster values based on a mapping.

        The raster must have an integer type.

        Parameters
        ----------
        remapping : str, dict
            Can be either a ``dict`` or a path string. If a ``dict`` is
            provided, the keys will be reclassified to the corresponding
            values. If a path string, it is treated as an ASCII remap file
            where each line looks like ``a:b`` and ``a`` and ``b`` are
            integers. All remap values (both from and to) must be integers.
        unmapped_to_null : bool, optional
            If ``True``, values not included in the mapping are instead mapped
            to the null value. Default is ``False``.

        Returns
        -------
        Raster
            The remapped raster.

        """
        from raster_tools.general import reclassify

        return reclassify(self, remapping, unmapped_to_null=unmapped_to_null)

    def round(self, decimals=0):
        """Evenly round to the given number of decimals

        Parameters
        ----------
        decimals : int, optional
            The number of decimal places to round to. If negative, value
            specifies the number of positions to the left of the decimal point.
            Default is 0.

        Returns
        -------
        Raster
            Rounded raster.

        """
        rast = self._ds.raster
        if self._masked:
            rast = xr.where(
                self._ds.mask, self.null_value, rast.round(decimals=decimals)
            ).rio.write_nodata(self.null_value)
        else:
            rast = rast.round(decimals=decimals)
        ds = make_raster_ds(rast, self._ds.mask)
        if self.crs is not None:
            ds = ds.rio.write_crs(self.crs)
        return Raster(ds, _fast_path=True)

    def xy(self, row, col, offset="center"):
        """Return the `(x, y)` coordinates of the pixel at `(row, col)`.

        Parameters
        ----------
        row : int, float, array-like
            row value(s) in the raster's CRS
        col : int, float, array-like
            row value(s) in the raster's CRS
        offset : str, optional
            Determines if the returned coordinates are for the center or a
            corner. Default is `'center'`

            'center'
                The pixel center.
            'ul'
                The upper left corner.
            'ur'
                The upper right corner.
            'll'
                The lower left corner.
            'lr'
                The lower right corner.

        Returns
        -------
        tuple
            (x value(s), y value(s)). If the inputs where array-like, the
            output values will be arrays.

        """
        return rowcol_to_xy(row, col, self.affine, offset)

    def index(self, x, y):
        """Return the `(row, col)` index of the pixel at `(x, y)`.

        Parameters
        ----------
        x : float, array-like
            x value(s) in the raster's CRS
        y : float, array-like
            y value(s) in the raster's CRS

        Returns
        -------
        tuple
            (row value(s), col value(s)). If the inputs where array-like, the
            output values will be arrays.

        """
        return xy_to_rowcol(x, y, self.affine)

    def to_vector(self):
        """
        Convert the raster into a vector where each non-null cell is a point.

        The resulting points are located at the center of each grid cell.

        Returns
        -------
        result
            The result is a dask_geopandas GeoDataFrame with the following
            columns: value, row, col, geometry. The value is the cell value
            at the row and column in the raster.

        """
        xrs = self._ds.raster
        mask = self._ds.mask.data
        data_delayed = xrs.data.to_delayed()
        chunks_shape = data_delayed.shape
        data_delayed = data_delayed.ravel()
        mask_delayed = mask.to_delayed().ravel()

        xc = da.from_array(xrs.x.data, chunks=xrs.chunks[2])
        xc_delayed = xc.to_delayed()
        yc = da.from_array(xrs.y.data, chunks=xrs.chunks[1])
        yc_delayed = yc.to_delayed()
        # See issue for inspiration: https://github.com/dask/dask/issues/7589
        # Group chunk data with corresponding coordinate data
        chunks = []
        for k, (d, m) in enumerate(zip(data_delayed, mask_delayed)):
            # band index, chunk y index, chunk x index
            b, i, j = np.unravel_index(k, chunks_shape)
            xck = xc_delayed[j]
            yck = yc_delayed[i]
            # Pass affine as tuple because of how dask handles namedtuple
            # subclasses.
            # ref: https://github.com/rasterio/affine/issues/79
            chunks.append(
                (d, m, xck, yck, b + 1, self.crs, tuple(self.affine))
            )

        meta = gpd.GeoDataFrame(
            {
                "value": [self.dtype.type(1)],
                "band": [1],
                "row": [0],
                "col": [0],
                "geometry": [Point(xrs.x.data[0], xrs.y.data[0])],
            },
            crs=self.crs,
        )
        results = [
            dd.from_delayed(_vectorize(*chunk), meta=meta) for chunk in chunks
        ]
        return dd.concat(results)

    def to_quadrants(self):
        """Split the raster into quadrants

        This returns the quadrants of the raster in the order northwest,
        northeast, southwest, southeast.

        Returns
        -------
        result : RasterQuadrantsResult
            The returned result is a `namedtuple` with attributes: `nw`, `ne`,
            `sw`, and `se`. Unpacking or indexing the object provides the
            quadrants in the stated order.

        """
        _, ny, nx = self.shape
        slice_nw = np.s_[:, : ny // 2, : nx // 2]
        slice_ne = np.s_[:, : ny // 2, nx // 2 :]
        slice_sw = np.s_[:, ny // 2 :, : nx // 2]
        slice_se = np.s_[:, ny // 2 :, nx // 2 :]
        slices = [slice_nw, slice_ne, slice_sw, slice_se]
        data = self.data.copy()
        mask = self.mask.copy()
        results = []
        for s in slices:
            data_quad = data[s]
            mask_quad = mask[s]
            x_quad = self.x[s[2]]
            y_quad = self.y[s[1]]
            xdata_quad = xr.DataArray(
                data_quad,
                coords=(self.band, y_quad, x_quad),
                dims=("band", "y", "x"),
            ).rio.write_nodata(self.null_value)
            xmask_quad = xr.DataArray(
                mask_quad,
                coords=(self.band, y_quad, x_quad),
                dims=("band", "y", "x"),
            )
            ds = make_raster_ds(xdata_quad, xmask_quad)
            if self.crs is not None:
                ds = ds.rio.write_crs(self.crs)
            results.append(Raster(ds, _fast_path=True))
        return RasterQuadrantsResult(*results)


_offset_name_to_rc_offset = {
    "center": (0.5, 0.5),
    "ul": (0, 0),
    "ur": (0, 1),
    "ll": (1, 0),
    "lr": (1, 1),
}


def rowcol_to_xy(row, col, affine, offset):
    """
    Convert (row, col) index values to (x, y) coords using the transformation.
    """
    roffset, coffset = _offset_name_to_rc_offset[offset]
    T = Affine.identity().translation(coffset, roffset)
    return affine * T * (col, row)


def xy_to_rowcol(x, y, affine):
    """
    Convert (x, y) coords to (row, col) index values using the transformation.
    """
    col, row = (~affine) * (x, y)
    row = np.floor(row).astype(int)
    col = np.floor(col).astype(int)
    return row, col


@jit(nopython=True, nogil=True)
def _extract_points(mask, xc, yc):
    n = mask.size - np.sum(mask)
    rx = np.empty(n, dtype=xc.dtype)
    ry = np.empty(n, dtype=yc.dtype)
    if n == 0:
        return rx, ry

    k = 0
    for i in range(mask.shape[1]):
        for j in range(mask.shape[2]):
            if mask[0, i, j]:
                continue
            rx[k] = xc[j]
            ry[k] = yc[i]
            k += 1
    return (rx, ry)


@jit(nopython=True, nogil=True)
def _extract_values(data, mask):
    n = mask.size - np.sum(mask)
    results = np.empty(n, dtype=data.dtype)
    if n == 0:
        return results

    k = 0
    for i in range(data.shape[1]):
        for j in range(data.shape[2]):
            if mask[0, i, j]:
                continue
            results[k] = data[0, i, j]
            k += 1
    return results


@dask.delayed
def _vectorize(data, mask, xc, yc, band, crs, affine_tuple):
    affine = Affine(*affine_tuple[:6])
    xpoints, ypoints = _extract_points(mask, xc, yc)
    if len(xpoints):
        values = _extract_values(data, mask)
        points = gpd.GeoSeries.from_xy(xpoints, ypoints, crs=crs)
        rows, cols = xy_to_rowcol(xpoints, ypoints, affine)
        bands = [band] * len(values)
    else:
        values = []
        points = gpd.GeoSeries([], crs=crs)
        bands = []
        rows = []
        cols = []
    df = gpd.GeoDataFrame(
        {
            "value": values,
            "band": bands,
            "row": rows,
            "col": cols,
            "geometry": points,
        },
        crs=crs,
    )
    if len(xpoints):
        return df
    # Add astype() as a workaround for
    # https://github.com/geopandas/dask-geopandas/issues/190
    return df.astype(
        {"value": data.dtype, "band": int, "row": int, "col": int}
    )


def get_raster(src, strict=True, null_to_nan=False):
    rs = None
    if isinstance(src, Raster):
        rs = src
    elif is_str(src):
        rs = Raster(src)
    elif strict:
        raise TypeError(
            f"Input must be a Raster or path string. Got {repr(type(src))}"
        )
    else:
        try:
            rs = Raster(src)
        except (ValueError, TypeError, RasterDataError, RasterIOError):
            raise ValueError(f"Could not convert input to Raster: {repr(src)}")

    if null_to_nan and rs._masked:
        rs = rs.copy()
        data = rs.data
        new_dtype = promote_dtype_to_float(data.dtype)
        if new_dtype != data.dtype:
            data = data.astype(new_dtype)
        rs.xdata.data = da.where(rs.mask, np.nan, data)
    return rs



import os

from setuptools import find_packages, setup


def read_file(fname):
    with open(fname, encoding="utf-8") as fd:
        return fd.read()


def get_version():
    with open(os.path.join("raster_tools", "_version.py")) as fd:
        # contents are __version__ = "<vstring>"
        return fd.read().split("=")[1].strip().strip('"')


def get_requirements(fname):
    with open(fname, encoding="utf-8") as fd:
        reqs = [line.strip() for line in fd if line]
    return reqs


NAME = "raster-tools"
DESCRIPTION = "Tools for processing geospatial data"
LONG_DESCRIPTION = read_file("README.md")
VERSION = get_version()
LICENSE = "GPL-3.0"
URL = "https://github.com/UM-RMRS/raster_tools"
PROJECT_URLS = {
    "Bug Tracker": "https://github.com/UM-RMRS/raster_tools/issues"
}
INSTALL_REQUIRES = get_requirements("requirements/default.txt")


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license=LICENSE,
    url=URL,
    project_urls=PROJECT_URLS,
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: GIS",
    ],
    package_dir={"": "."},
    packages=find_packages(exclude=["docs", "tests"]),
    install_requires=INSTALL_REQUIRES,
    python_requires=">=3.8",
)



import unittest
from functools import partial

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from scipy.ndimage import (
    binary_dilation,
    binary_erosion,
    grey_dilation,
    grey_erosion,
)

from raster_tools import creation, general
from raster_tools.dtypes import F32, F64, I16, I32, I64, U8, U16, is_scalar
from raster_tools.masking import get_default_null_value
from raster_tools.raster import Raster, get_raster
from raster_tools.stat_common import (
    nan_unique_count_jit,
    nanargmax_jit,
    nanargmin_jit,
    nanasm_jit,
    nanentropy_jit,
    nanmode_jit,
)
from tests.utils import assert_valid_raster

stat_funcs = {
    "max": partial(np.nanmax, axis=0),
    "mean": partial(np.nanmean, axis=0),
    "median": partial(np.nanmedian, axis=0),
    "min": partial(np.nanmin, axis=0),
    "prod": partial(np.nanprod, axis=0),
    "std": partial(np.nanstd, axis=0),
    "sum": partial(np.nansum, axis=0),
    "var": partial(np.nanvar, axis=0),
}
custom_stat_funcs = {
    "asm": nanasm_jit,
    "entropy": nanentropy_jit,
    "maxband": nanargmax_jit,
    "minband": nanargmin_jit,
    "mode": nanmode_jit,
    "unique": nan_unique_count_jit,
}


def get_local_stats_dtype(stat, input_data):
    if stat == "unique":
        dt = np.min_scalar_type(input_data.shape[0])
        if dt == U8:
            return I16
        if dt == U16:
            return I32
    if stat in ("mode", "min", "max"):
        return input_data.dtype
    if stat in ("minband", "maxband"):
        dt = np.min_scalar_type(input_data.shape[0] - 1)
        if dt == U8:
            return I16
        if dt == U16:
            return I32
    if input_data.dtype == F32:
        return F32
    return F64


@pytest.mark.parametrize("chunk", [False, True])
@pytest.mark.parametrize("stat", list(custom_stat_funcs.keys()))
def test_local_stats(stat, chunk):
    for dt in (I32, I64, F32, F64):
        x = np.arange(5 * 4 * 4).reshape(5, 4, 4) - 20
        x[2, :, :-1] = 1
        x[:, 2, 2] = 1
        rs = Raster(x.astype(dt)).set_null_value(1)
        if chunk:
            rs = rs.chunk((1, 2, 2))
            orig_chunks = rs.data.chunks
        xx = np.where(rs.mask.compute(), np.nan, rs.values)

        if stat in stat_funcs:
            sfunc = stat_funcs[stat]
            truth = sfunc(xx)[None]
            if stat != "sum":
                truth = np.where(np.isnan(truth), 1, truth)
            else:
                truth = np.where(
                    np.all(rs.mask, axis=0).compute(), rs.null_value, truth
                )
            result = general.local_stats(rs, stat)
        else:
            sfunc = custom_stat_funcs[stat]
            truth = np.zeros(
                (1, *rs.shape[1:]), dtype=get_local_stats_dtype(stat, rs.data)
            )
            for i in range(rs.shape[1]):
                for j in range(rs.shape[2]):
                    v = sfunc(xx[:, i, j])
                    if np.isnan(v):
                        v = rs.null_value
                    truth[0, i, j] = v
            if stat in ("unique", "minband", "maxband"):
                nv = get_default_null_value(I16)
                truth = np.where(
                    np.all(rs.mask, axis=0).compute(), nv, truth.astype(I16)
                )
            else:
                truth = np.where(
                    np.all(rs.mask, axis=0).compute(),
                    get_default_null_value(truth.dtype),
                    truth,
                )
            result = general.local_stats(rs, stat)
        assert_valid_raster(result)
        assert result.shape[0] == 1
        assert result.shape == truth.shape
        assert np.allclose(result, truth, equal_nan=True)
        if chunk:
            assert result.data.chunks == ((1,), *orig_chunks[1:])
        else:
            assert result.data.chunks == ((1,), *rs.data.chunks[1:])
        assert result.dtype == get_local_stats_dtype(stat, rs.data)
        assert result.crs == rs.crs
        assert result.affine == rs.affine


def test_local_stats_reject_bad_stat():
    rs = Raster(np.arange(5 * 4 * 4).reshape(5, 4, 4))

    for stat in [0, np.nanvar, float]:
        with pytest.raises(TypeError):
            general.local_stats(rs, stat)
    for stat in ["", "nanstd", "minn"]:
        with pytest.raises(ValueError):
            general.local_stats(rs, stat)


coarsen_stats = {
    "max": lambda x: x.max(),
    "mean": lambda x: x.mean(),
    "median": lambda x: x.median(),
    "min": lambda x: x.min(),
    "prod": lambda x: x.prod(),
    "std": lambda x: x.std(),
    "sum": lambda x: x.sum(),
    "var": lambda x: x.var(),
}
coarsen_custom_stats = {
    "asm": nanasm_jit,
    "entropy": nanentropy_jit,
    "mode": nanmode_jit,
    "unique": nan_unique_count_jit,
}


def get_coarsen_dtype(stat, window_size, input_data):
    if stat == "unique":
        return np.min_scalar_type(window_size)
    if stat in ("mode", "min", "max"):
        return input_data.dtype
    if input_data.dtype == F32:
        return F32
    return F64


def coarsen_block(x, axis, func, out_dtype):
    shape = tuple(d for i, d in enumerate(x.shape) if i not in axis)
    out = np.empty(shape, dtype=out_dtype)
    for i in range(shape[0]):
        for j in range(shape[1]):
            for k in range(shape[2]):
                v = func(x[i, j, :, k, :])
                out[i, j, k] = 0 if np.isnan(v) else v
    return out


def coarsen_reduce(x, axis, agg_func, out_dtype):
    dims = tuple(i for i in range(len(x.shape)) if i not in axis)
    chunks = tuple(x.chunks[d] for d in dims)
    return x.map_blocks(
        partial(coarsen_block, axis=axis, func=agg_func, out_dtype=out_dtype),
        chunks=chunks,
        drop_axis=axis,
        meta=np.array((), dtype=out_dtype),
    )


@pytest.mark.parametrize("chunk", [False, True])
@pytest.mark.parametrize("window_x", [1, 3])
@pytest.mark.parametrize("window_y", [2, 3, 4])
@pytest.mark.parametrize("stat", list(coarsen_custom_stats.keys()))
def test_aggregate(stat, window_y, window_x, chunk):
    window = (window_y, window_x)
    window_map = {"y": window_y, "x": window_x}
    x = np.arange(4 * 11 * 11).reshape((4, 11, 11)) - 20
    null_value = -999
    x[1, 6, :] = null_value
    x[2, 10, :] = null_value
    x[3, :, :] = null_value
    rs = Raster(x).set_null_value(null_value)
    rs._ds = rs._ds.rio.write_crs("EPSG:3857")
    if chunk:
        rs = rs.chunk((1, 3, 3))
    xmask = rs.xmask
    xmask_truth = xmask.coarsen(dim=window_map, boundary="trim").all()

    if stat in coarsen_stats:
        stat_func = coarsen_stats[stat]
        xtruth = stat_func(
            get_raster(rs, null_to_nan=True).xdata.coarsen(
                dim=window_map, boundary="trim"
            )
        )
        xtruth = (
            xr.where(xmask_truth, null_value, xtruth)
            .astype(get_coarsen_dtype(stat, np.prod(window), rs.xdata))
            .rio.write_crs(rs.crs)
        )
        truth = Raster(xtruth).set_null_value(null_value)
    else:
        stat_func = coarsen_custom_stats[stat]
        out_dtype = get_coarsen_dtype(stat, np.prod(window), rs.xdata)
        xtruth = get_raster(rs, null_to_nan=True).xdata
        xtruth = (
            xtruth.coarsen(dim=window_map, boundary="trim")
            .reduce(
                partial(
                    coarsen_reduce,
                    agg_func=stat_func,
                    out_dtype=out_dtype,
                )
            )
            .compute()
        )
        if stat == "unique":
            nv = get_default_null_value(I16)
            xtruth = xr.where(
                xmask_truth, nv, xtruth.astype(I16)
            ).rio.write_crs(rs.crs)
            truth = Raster(xtruth).set_null_value(nv)
        else:
            xtruth = xr.where(
                xmask_truth, get_default_null_value(xtruth.dtype), xtruth
            ).rio.write_crs(rs.crs)
            assert xtruth.dtype == np.promote_types(
                np.min_scalar_type(null_value), out_dtype
            )
            truth = Raster(xtruth).set_null_value(null_value)
        truth.mask.data = xmask.data
    result = general.aggregate(rs, window, stat)

    assert_valid_raster(result)
    assert xtruth.equals(result.xdata)
    assert rs.crs == result.crs
    assert truth.affine == result.affine
    assert all(
        result.resolution == np.atleast_1d(rs.resolution) * window[::-1]
    )
    assert result.null_value == nv if stat == "unique" else null_value


@pytest.mark.parametrize(
    "window,stat,error_type",
    [
        ((1, 1), "mean", ValueError),
        ((3, 0), "mean", ValueError),
        ((1, 2, 3), "mean", ValueError),
        ((1.0, 1), "mean", TypeError),
        ((3, 3), "other", ValueError),
        ((3, 3), np.sum, TypeError),
    ],
)
def test_aggregate_errors(window, stat, error_type):
    rs = Raster("tests/data/raster/elevation.tif")
    with pytest.raises(error_type):
        general.aggregate(rs, window, stat)


@pytest.mark.parametrize("chunk", [False, True])
@pytest.mark.parametrize("null_value", [-99, 100, np.nan])
@pytest.mark.parametrize(
    "size", [2, 3, 4, 5, (3, 1), (1, 4), (2, 3), (3, 3), (3, 5), (4, 6)]
)
@pytest.mark.parametrize("name", ["erode", "dilate"])
def test_erode_dilate(name, size, null_value, chunk):
    erode_or_dilate = getattr(general, name)
    grey_erode_or_dilate = grey_erosion if name == "erode" else grey_dilation
    binary_erode_or_dilate = (
        binary_erosion if name == "erode" else binary_dilation
    )

    nan_null = np.isnan(null_value)
    x = np.full((15, 15), null_value)
    x[4:10, 4:11] = 2
    x[5:8, 6:10] = 3
    x[:, 3] = 1
    x[5] = 4
    x[12, 12] = 3
    x = np.stack((x, x))
    if nan_null:
        mask = np.isnan(x)
    else:
        mask = x == null_value
    x[1] += 1
    if not nan_null:
        x[mask] = null_value
    rs = Raster(x).set_null_value(null_value)
    if chunk:
        rs = rs.chunk((1, 4, 4))
    rs._ds = rs._ds.rio.write_crs("EPSG:3857")
    tup_size = size if isinstance(size, tuple) else (size, size)

    ntruth = x.copy()
    mask_truth = mask.copy()
    fill = 0
    ntruth[mask] = fill
    for bnd in range(len(ntruth)):
        ntruth[bnd] = grey_erode_or_dilate(
            ntruth[bnd], size=size, mode="constant", cval=fill
        )
        mask_truth[bnd] = ~binary_erode_or_dilate(
            ~mask_truth[bnd], structure=np.ones(tup_size, dtype=bool)
        )
    ntruth[mask_truth] = null_value
    xtruth = xr.zeros_like(rs.xdata)
    xtruth.data = ntruth
    truth = Raster(xtruth).set_null_value(null_value)
    truth._ds.mask.data = da.from_array(mask_truth)

    result = erode_or_dilate(rs, size)

    assert result.dtype == rs.dtype
    assert xtruth.equals(result.xdata)
    assert np.allclose(result, truth, equal_nan=True)
    assert np.allclose(result.mask.compute(), mask_truth)
    assert rs.crs == result.crs
    if not nan_null:
        assert rs.null_value == result.null_value
    else:
        assert np.isnan(result.null_value)


@pytest.mark.parametrize("name", ["erode", "dilate"])
def test_erode_dilate_errors(name):
    func = getattr(general, name)
    rs = Raster("tests/data/raster/elevation_small.tif")

    for size in [3.0, None]:
        with pytest.raises(TypeError):
            func(rs, size)
    for size in [0, 1, -1, (3,), (1, 1), (3, 3, 3), (-3, 3), ()]:
        with pytest.raises(ValueError):
            func(rs, size)


# TODO: fully test module
class TestSurface(unittest.TestCase):
    def setUp(self):
        self.dem = Raster("tests/data/raster/elevation.tif")
        self.multi = Raster("tests/data/raster/multiband_small.tif")

    def test_regions(self):
        rs_pos = creation.random_raster(
            self.dem, distribution="poisson", bands=1, params=[7, 0.5]
        )
        r = general.regions(rs_pos)
        assert_valid_raster(r)
        r.eval()


def test_band_concat():
    rs1 = Raster("tests/data/raster/elevation_clipped_small.tif")
    rs2 = rs1 + 30
    rsnp1 = rs1.values
    rsnp2 = rs2.values
    mask1 = rs1.xmask.values
    mask2 = rs2.xmask.values
    truth = np.concatenate((rsnp1, rsnp2))
    mask = np.concatenate((mask1, mask2))
    truth[mask] = get_default_null_value(truth.dtype)
    assert truth.ndim == 3
    test = general.band_concat([rs1, rs2])
    assert_valid_raster(test)
    assert test.crs == rs1.crs
    assert test.dtype == truth.dtype
    assert test.null_value == get_default_null_value(test.dtype)
    assert test.shape == truth.shape
    assert np.allclose(test, truth)
    truth = np.concatenate((rsnp1, rsnp1, rsnp2, truth))
    mask = np.concatenate((mask1, mask1, mask2, mask))
    truth[mask] = get_default_null_value(truth.dtype)
    test = general.band_concat([rs1, rs1, rs2, test])
    assert_valid_raster(test)
    assert test.crs == rs1.crs
    assert test.dtype == truth.dtype
    assert test.null_value == get_default_null_value(test.dtype)
    assert test.shape == truth.shape
    assert np.allclose(test, truth)


@pytest.mark.filterwarnings("ignore:The null value")
def test_band_concat_bool_rasters():
    rs1 = Raster("tests/data/raster/elevation_small.tif") > -100
    rs2 = rs1.copy()
    result = general.band_concat((rs1, rs2))
    assert result.null_value == get_default_null_value(result.dtype)
    assert result.dtype == np.dtype(bool)
    assert np.array(result).all()

    result = general.band_concat((rs1, rs2))
    # The null value will be changed to match type
    assert result.null_value == get_default_null_value(bool)
    assert result.dtype == np.dtype("?")
    assert np.all(np.array(result) == 1)


def test_band_concat_errors():
    with pytest.raises(ValueError):
        general.band_concat([])


def test_remap_range():
    rs = Raster(np.arange(25).reshape((5, 5)))
    rsnp = rs.values

    mapping = (0, 5, 1)
    result = general.remap_range(rs, mapping)
    truth = rsnp.copy()
    truth[(rsnp >= mapping[0]) & (rsnp < mapping[1])] = mapping[2]
    assert np.allclose(result, truth)
    assert np.allclose(rs.remap_range(mapping), truth)

    mappings = [mapping, (5, 15, -1)]
    result = general.remap_range(rs, mappings)
    truth[(rsnp >= mappings[1][0]) & (rsnp < mappings[1][1])] = mappings[1][2]
    assert np.allclose(result, truth)
    assert np.allclose(rs.remap_range(mappings), truth)

    # Test multiple with potential conflict in last 2
    mappings = [(0, 1, 0), (1, 2, 1), (2, 3, 8), (8, 9, 2)]
    result = general.remap_range(rs, mappings)
    truth = rsnp.copy()
    for m in mappings:
        truth[(rsnp >= m[0]) & (rsnp < m[1])] = m[2]
    assert np.allclose(result.values, truth)
    assert np.allclose(rs.remap_range(mappings), truth)

    # Test precedence
    mappings = [(0, 2, 0), (1, 2, 1)]
    result = general.remap_range(rs, mappings)
    truth = rsnp.copy()
    m = mappings[0]
    truth[(rsnp >= m[0]) & (rsnp < m[1])] = m[2]
    assert np.allclose(result.values, truth)
    assert np.allclose(rs.remap_range(mappings), truth)


@pytest.mark.parametrize(
    "rast,mapping",
    [
        (Raster(np.arange(16).reshape((4, 4))), [(0, 4, -1), (8, 12, 0)]),
        (
            Raster(np.arange(32).reshape((2, 4, 4))),
            [(0, 4, -1), (8, 12, 0), (15, 20, 0)],
        ),
    ],
)
@pytest.mark.parametrize("inc", ["left", "right", "both", "none"])
def test_remap_range_inclusivity(rast, mapping, inc):
    data = rast.values.copy()
    hist = np.zeros(data.shape, dtype=bool)
    for left, right, new in mapping:
        if inc == "left":
            mask = (left <= data) & (data < right)
        elif inc == "right":
            mask = (left < data) & (data <= right)
        elif inc == "both":
            mask = (left <= data) & (data <= right)
        else:
            mask = (left < data) & (data < right)
        mask[hist] = 0
        if mask.any():
            data[mask] = new
        hist |= mask

    result = general.remap_range(rast, mapping, inc)

    assert np.allclose(result, data)
    assert np.allclose(rast.remap_range(mapping, inc), data)


def test_remap_range_f16():
    rs = Raster(np.arange(25).reshape((5, 5))).astype("float16")
    rsnp = rs.values
    mapping = (0, 5, 1)
    result = general.remap_range(rs, mapping)
    truth = rsnp.copy()
    truth[(rsnp >= mapping[0]) & (rsnp < mapping[1])] = mapping[2]
    assert rs.dtype == np.dtype("float16")
    assert result.dtype == np.dtype("float16")
    assert np.allclose(result, truth)

    rs = Raster(np.arange(25).reshape((5, 5))).astype("int8")
    rsnp = rs.values
    mapping = (0, 5, 2.0)
    result = general.remap_range(rs, mapping)
    truth = rsnp.copy().astype("float16")
    truth[(rsnp >= mapping[0]) & (rsnp < mapping[1])] = mapping[2]
    assert rs.dtype == np.dtype("int8")
    assert result.dtype == np.dtype("float16")
    assert np.allclose(result, truth)


def test_remap_range_errors():
    rs = Raster("tests/data/raster/elevation_small.tif")
    # TypeError if not scalars
    with pytest.raises(TypeError):
        general.remap_range(rs, (None, 2, 4))
    with pytest.raises(TypeError):
        general.remap_range(rs, (0, "2", 4))
    with pytest.raises(TypeError):
        general.remap_range(rs, (0, 2, None))
    with pytest.raises(TypeError):
        general.remap_range(rs, [(0, 2, 1), (2, 3, None)])
    # ValueError if nan
    with pytest.raises(ValueError):
        general.remap_range(rs, (np.nan, 2, 4))
    with pytest.raises(ValueError):
        general.remap_range(rs, (0, np.nan, 4))
    # ValueError if range reversed
    with pytest.raises(ValueError):
        general.remap_range(rs, (0, -1, 6))
    with pytest.raises(ValueError):
        general.remap_range(rs, (1, 1, 6))
    with pytest.raises(ValueError):
        general.remap_range(rs, (0, 1))
    with pytest.raises(ValueError):
        general.remap_range(rs, [(0, 1, 2), (0, 3)])
    with pytest.raises(ValueError):
        general.remap_range(rs, ())


def get_where_truth(cond, x, y):
    if cond.dtype != np.dtype(bool):
        cond = cond > 0

    chunks = "auto"
    if not is_scalar(x):
        chunks = x.data.chunks
    elif not is_scalar(y):
        chunks = y.data.chunks
    cond = np.array(cond)
    nx = (
        np.ma.array(x.data.compute(), mask=x.mask.compute())
        if not is_scalar(x)
        else x
    )
    ny = (
        np.ma.array(y.data.compute(), mask=y.mask.compute())
        if not is_scalar(y)
        else y
    )
    masked = any(r._masked if isinstance(r, Raster) else False for r in (x, y))
    scalar_and_nan = (
        all(is_scalar(r) for r in (x, y)) and np.isnan((x, y)).any()
    )
    masked |= scalar_and_nan

    xmask = x.mask.compute() if not is_scalar(x) else np.zeros_like(cond)
    ymask = y.mask.compute() if not is_scalar(y) else np.zeros_like(cond)

    result = np.ma.where(cond, nx, ny)
    if scalar_and_nan:
        mask = np.isnan(result)
    else:
        mask = np.where(cond, xmask, ymask)
    result = Raster(np.array(result))
    if chunks != "auto":
        result = result.chunk(chunks)
    if masked:
        result.xmask.data = da.from_array(mask, chunks=chunks)
        result._ds["raster"] = result.xdata.rio.write_nodata(
            get_default_null_value(result.dtype)
        )
        result = result.burn_mask()
    return result


wshape = (10, 10)
nwhere = np.prod(wshape)


def create_rs(x):
    x = Raster(x)
    x._ds = x._ds.rio.write_crs("EPSG:3857")
    return x


@pytest.mark.parametrize(
    "x,y",
    [
        (0, 1),
        (0, np.nan),
        (0, create_rs(np.ones(wshape))),
        (0, create_rs(np.ones(wshape)).set_null_value(1)),
        (create_rs(np.ones(wshape)), 0),
        (create_rs(np.zeros(wshape)), create_rs(np.ones(wshape))),
        (
            create_rs(np.zeros(wshape)).set_null_value(0),
            create_rs(np.ones(wshape)).set_null_value(1),
        ),
        (
            create_rs(np.arange(nwhere).reshape(wshape)).set_null_value(10),
            create_rs(np.arange(nwhere).reshape(wshape) + 20).set_null_value(
                45
            ),
        ),
    ],
)
@pytest.mark.parametrize(
    "cond",
    [
        create_rs(np.zeros(wshape, dtype=bool)),
        create_rs(np.ones(wshape, dtype=bool)),
        create_rs(np.arange(nwhere).reshape(wshape) > 40),
        create_rs(np.arange(nwhere).reshape(wshape) % 10),
        create_rs((np.arange(nwhere).reshape(wshape) % 10) > 0),
        create_rs((np.arange(nwhere).reshape(wshape) % 4) > 0),
    ],
)
def test_where(cond, x, y):
    truth = get_where_truth(cond, x, y)
    result = general.where(cond, x, y)

    assert_valid_raster(result)
    assert np.allclose(result, truth, equal_nan=True)
    assert result.dtype == result.dtype
    if result._masked:
        assert not np.isnan(result.null_value)
    assert result.null_value == truth.null_value
    assert np.allclose(result.xmask, truth.xmask)
    assert result.data.chunks == truth.data.chunks
    assert result.crs is not None
    assert result.crs == "EPSG:3857"


@pytest.mark.parametrize("method", [False, True])
def test_reclassify(method):
    data = np.arange(100).reshape((10, 10))
    rast = Raster(Raster(data)._ds.rio.write_crs("EPSG:3857"))
    mapping = {k: v for k, v in zip(np.arange(40), np.arange(20, 60))}
    included_mask = np.array([v in mapping for v in data.ravel()]).reshape(
        (10, 10)
    )

    truth = data.copy()
    truth[included_mask] = np.arange(20, 60)
    result = (
        rast.reclassify(mapping)
        if method
        else general.reclassify(rast, mapping)
    )
    assert np.allclose(result, truth)
    assert result.crs == rast.crs

    truth = data.copy()
    truth[included_mask] = np.arange(20, 60)
    truth[~included_mask] = get_default_null_value(int)
    result = (
        rast.reclassify(mapping, True)
        if method
        else general.reclassify(rast, mapping, True)
    )
    assert np.allclose(general.reclassify(rast, mapping, True), truth)
    assert result.null_value == get_default_null_value(int)
    assert result.crs == rast.crs

    rast = rast.set_null_value(-1)
    truth = data.copy()
    truth[included_mask] = np.arange(20, 60)
    truth[~included_mask] = -1
    result = (
        rast.reclassify(mapping, True)
        if method
        else general.reclassify(rast, mapping, True)
    )
    assert np.allclose(result, truth)
    assert result.null_value == -1
    assert result.crs == rast.crs



import dask.array as da
import numba as nb
import numpy as np
import xarray as xr

from raster_tools.dtypes import F32, F64
from raster_tools.masking import get_default_null_value
from raster_tools.raster import Raster, get_raster
from raster_tools.utils import single_band_mappable

__all__ = [
    "pa_allocation",
    "pa_direction",
    "pa_proximity",
    "proximity_analysis",
]

_JIT_KWARGS = {"nopython": True, "nogil": True}
ngjit = nb.jit(**_JIT_KWARGS)


@nb.jit(
    [
        "float64(float64, float64, float64, float64)",
        "int64(int64, int64, int64, int64)",
    ],
    **_JIT_KWARGS,
)
def _euclidean_dist_sqr(x1, y1, x2, y2):
    dx = x1 - x2
    dy = y1 - y2
    return (dx * dx) + (dy * dy)


@nb.jit(
    [
        "float64(float64, float64, float64, float64)",
        "int64(int64, int64, int64, int64)",
    ],
    **_JIT_KWARGS,
)
def _taxi_cab_dist_sqr(x1, y1, x2, y2):
    dx = x1 - x2
    dy = y1 - y2
    dist = np.abs(dx) + np.abs(dy)
    return dist * dist


@nb.jit(
    [
        "float64(float64, float64, float64, float64)",
        "float64(int64, int64, int64, int64)",
    ],
    **_JIT_KWARGS,
)
def _chebyshev_dist_sqr(x1, y1, x2, y2):
    dx = np.abs(x1 - x2)
    dy = np.abs(y1 - y2)
    dist = max(dx, dy)
    return dist * dist


@nb.jit(
    [
        "float64(float64, float64, float64, float64)",
        "float64(int64, int64, int64, int64)",
    ],
    **_JIT_KWARGS,
)
def _haversine_dist_sqr(x1, y1, x2, y2):
    # Great Circle or Haversine distance in meters (squared)
    # ref: https://en.wikipedia.org/wiki/Great-circle_distance
    # ref: https://en.wikipedia.org/wiki/Haversine_formula
    x1 = np.radians(x1)
    y1 = np.radians(y1)
    x2 = np.radians(x2)
    y2 = np.radians(y2)
    sinlat = np.sin((y2 - y1) / 2.0)
    sinlon = np.sin((x2 - x1) / 2.0)
    h = (sinlat * sinlat) + (np.cos(y1) * np.cos(y2) * sinlon * sinlon)
    # Mean Earth radius for WGS84
    r = 6371009.0
    hav = 2 * r * np.arcsin(np.sqrt(h))
    return hav * hav


_DISTANCE_FUNC_MAP = {
    # Euclidean (default)
    None: _euclidean_dist_sqr,
    "euclidean": _euclidean_dist_sqr,
    # Taxi
    "taxi": _taxi_cab_dist_sqr,
    "taxicab": _taxi_cab_dist_sqr,
    "manhatten": _taxi_cab_dist_sqr,
    "city_block": _taxi_cab_dist_sqr,
    # Chebyshev
    "chebyshev": _chebyshev_dist_sqr,
    "chessboard": _chebyshev_dist_sqr,
    # Great circle
    "haversine": _haversine_dist_sqr,
    "great_circle": _haversine_dist_sqr,
}


@nb.jit(
    [
        "float64(float64, float64, float64, float64)",
        "float64(int64, int64, int64, int64)",
    ],
    **_JIT_KWARGS,
)
def _direction(x1, y1, x2, y2):
    # (x1, y1): current cell
    # (x2, y2): nearest target cell
    # From ArcMap Euclidean Direction docs: The output values are based on
    # compass directions (90 to the east, 180 to the south, 270 to the west,
    # and 360 to the north), with 0 reserved for the source cells
    #            y,N,360
    #            ^
    #            |---.
    #            |    \
    #            |     `
    #            |     |
    # W,270 <----+-----v-> x,E,90
    #            |
    #            |
    #            v
    #            S,180
    if x1 == x2 and y1 == y2:
        return 0.0

    dx = x2 - x1
    dy = y2 - y1
    # -dy to flip so theta increases clockwise
    th = np.arctan2(-dy, dx) * 57.29577951308232
    # map to 0-360 range
    # +90 to shift so north is along y axis
    th = (th + 90.0 + 360) % 360
    if th == 0.0:
        th = 360.0
    return th


_MODE_PROXIMITY = 1
_MODE_ALLOCATION = 2
_MODE_DIRECTION = 3


@ngjit
def _process_proximity_line(
    scan_line,
    line_proximity,
    pan_near_xi,
    pan_near_yi,
    pan_nearest_xi,
    pan_nearest_yi,
    xc,
    yc,
    forward,
    iline,
    max_distance,
    targets,
    dist_sqr_func,
    mode,
):
    # Processes a single line in the direction determined by `forward` and
    # updates the proximities for the line. pan_near_{x, y} are buffers that
    # keep track of the closest points seen so far on this pass up or down the
    # raster.
    nx = len(scan_line)
    nt = len(targets)
    start = 0 if forward else nx - 1
    end = nx if forward else -1
    step = 1 if forward else -1

    max_dist_sqr = max_distance * max_distance
    max_dist_sqr_2 = max_dist_sqr * 2.0

    for ipixel in range(start, end, step):
        is_target = False
        if nt > 0:
            for j in range(nt):
                is_target |= scan_line[j] == targets[j]
        else:
            # No targets so assume all finite and non-zero pixels are targets
            is_target = (scan_line[ipixel] != 0) & np.isfinite(
                scan_line[ipixel]
            )
        # Are we on a target?
        if is_target:
            line_proximity[ipixel] = 0
            pan_near_xi[ipixel] = ipixel
            pan_near_yi[ipixel] = iline
            if mode != _MODE_PROXIMITY:
                pan_nearest_xi[ipixel] = ipixel
                pan_nearest_yi[ipixel] = iline
            continue

        # Are we near(er) to the closest target to the above/below pixel?
        near_dist_sqr = max_dist_sqr_2
        if pan_near_xi[ipixel] >= 0:
            dist_sqr = dist_sqr_func(
                xc[pan_near_xi[ipixel]],
                yc[pan_near_yi[ipixel]],
                xc[ipixel],
                yc[iline],
            )
            if dist_sqr < near_dist_sqr:
                near_dist_sqr = dist_sqr
            else:
                pan_near_xi[ipixel] = -1
                pan_near_yi[ipixel] = -1
        # Are we near(er) to the closest target in the backward diagonal
        # direction?
        iprev = ipixel - step
        if ipixel != start and pan_near_xi[iprev] >= 0:
            dist_sqr = dist_sqr_func(
                xc[pan_near_xi[iprev]],
                yc[pan_near_yi[iprev]],
                xc[ipixel],
                yc[iline],
            )
            if dist_sqr < near_dist_sqr:
                near_dist_sqr = dist_sqr
                pan_near_xi[ipixel] = pan_near_xi[iprev]
                pan_near_yi[ipixel] = pan_near_yi[iprev]
        # Are we near(er) to the closest target in the forward diagonal
        # direction?
        inext = ipixel + step
        if inext != end and pan_near_xi[inext] >= 0:
            dist_sqr = dist_sqr_func(
                xc[pan_near_xi[inext]],
                yc[pan_near_yi[inext]],
                xc[ipixel],
                yc[iline],
            )
            if dist_sqr < near_dist_sqr:
                near_dist_sqr = dist_sqr
                pan_near_xi[ipixel] = pan_near_xi[inext]
                pan_near_yi[ipixel] = pan_near_yi[inext]
        # Update proximity
        if (
            pan_near_xi[ipixel] >= 0
            and near_dist_sqr <= max_dist_sqr
            and (
                line_proximity[ipixel] < 0
                or near_dist_sqr < line_proximity[ipixel] ** 2
            )
        ):
            line_proximity[ipixel] = np.sqrt(near_dist_sqr)
            if mode != _MODE_PROXIMITY:
                pan_nearest_xi[ipixel] = pan_near_xi[ipixel]
                pan_nearest_yi[ipixel] = pan_near_yi[ipixel]


@ngjit
def _compute_proximity(
    src,
    proximity,
    secondary,
    target_values,
    xc,
    yc,
    max_distance,
    dist_sqr_func,
    nodata,
    mode,
):
    # This func works by making passes forward and backward (row axis) through
    # the raster, iteretively updating the proximities to targets. It is based
    # on GDAL's proximity implementation.
    #
    # ref: https://github.com/OSGeo/gdal/blob/master/alg/gdalproximity.cpp
    ny, nx = src.shape

    # Buffer for rows from raster
    scan_line = np.empty(nx, dtype=src.dtype)
    # Buffer for proximity values along rows
    line_proximity = np.empty(nx, dtype=proximity.dtype)
    # Arrays to keep track of nearest targets seen so far. Pan means global
    # (raster-chunk-wide), here
    pan_near_xi = np.empty(nx, dtype=np.int64)
    pan_near_yi = np.empty(nx, dtype=np.int64)
    # Only allocate full arrays if needed for secondary calculation
    if mode == _MODE_PROXIMITY:
        pan_nearest_xi = np.empty(0, dtype=np.int64)
        pan_nearest_yi = np.empty(0, dtype=np.int64)
    else:
        pan_nearest_xi = np.empty(nx, dtype=np.int64)
        pan_nearest_yi = np.empty(nx, dtype=np.int64)

    # Top to bottom
    pan_near_xi[:] = -1
    pan_near_yi[:] = -1
    for iline in range(ny):
        scan_line[:] = src[iline]
        line_proximity[:] = -1

        # Left to right
        pan_nearest_xi[:] = -1
        pan_nearest_yi[:] = -1
        _process_proximity_line(
            scan_line,
            line_proximity,
            pan_near_xi,
            pan_near_yi,
            pan_nearest_xi,
            pan_nearest_yi,
            xc,
            yc,
            True,
            iline,
            max_distance,
            target_values,
            dist_sqr_func,
            mode,
        )
        if mode != _MODE_PROXIMITY:
            for i in range(nx):
                if pan_nearest_xi[i] >= 0 and line_proximity[i] >= 0:
                    if mode == _MODE_ALLOCATION:
                        secondary[iline, i] = src[
                            pan_nearest_yi[i], pan_nearest_xi[i]
                        ]
                    else:
                        secondary[iline, i] = _direction(
                            xc[i],
                            yc[iline],
                            xc[pan_nearest_xi[i]],
                            yc[pan_nearest_yi[i]],
                        )
        # Right to left
        pan_nearest_xi[:] = -1
        pan_nearest_yi[:] = -1
        _process_proximity_line(
            scan_line,
            line_proximity,
            pan_near_xi,
            pan_near_yi,
            pan_nearest_xi,
            pan_nearest_yi,
            xc,
            yc,
            False,
            iline,
            max_distance,
            target_values,
            dist_sqr_func,
            mode,
        )
        proximity[iline] = line_proximity
        if mode != _MODE_PROXIMITY:
            for i in range(nx):
                if pan_nearest_xi[i] >= 0 and line_proximity[i] >= 0:
                    if mode == _MODE_ALLOCATION:
                        secondary[iline, i] = src[
                            pan_nearest_yi[i], pan_nearest_xi[i]
                        ]
                    else:
                        secondary[iline, i] = _direction(
                            xc[i],
                            yc[iline],
                            xc[pan_nearest_xi[i]],
                            yc[pan_nearest_yi[i]],
                        )

    # Bottom to top
    pan_near_xi[:] = -1
    pan_near_yi[:] = -1
    for iline in range(ny - 1, -1, -1):
        scan_line[:] = src[iline]
        line_proximity[:] = proximity[iline]

        # Right to left
        pan_nearest_xi[:] = -1
        pan_nearest_yi[:] = -1
        _process_proximity_line(
            scan_line,
            line_proximity,
            pan_near_xi,
            pan_near_yi,
            pan_nearest_xi,
            pan_nearest_yi,
            xc,
            yc,
            False,
            iline,
            max_distance,
            target_values,
            dist_sqr_func,
            mode,
        )
        if mode != _MODE_PROXIMITY:
            for i in range(nx):
                if pan_nearest_xi[i] >= 0 and line_proximity[i] >= 0:
                    if mode == _MODE_ALLOCATION:
                        secondary[iline, i] = src[
                            pan_nearest_yi[i], pan_nearest_xi[i]
                        ]
                    else:
                        secondary[iline, i] = _direction(
                            xc[i],
                            yc[iline],
                            xc[pan_nearest_xi[i]],
                            yc[pan_nearest_yi[i]],
                        )
        # Left to right
        pan_nearest_xi[:] = -1
        pan_nearest_yi[:] = -1
        _process_proximity_line(
            scan_line,
            line_proximity,
            pan_near_xi,
            pan_near_yi,
            pan_nearest_xi,
            pan_nearest_yi,
            xc,
            yc,
            True,
            iline,
            max_distance,
            target_values,
            dist_sqr_func,
            mode,
        )
        if mode != _MODE_PROXIMITY:
            for i in range(nx):
                if pan_nearest_xi[i] >= 0 and line_proximity[i] >= 0:
                    if mode == _MODE_ALLOCATION:
                        secondary[iline, i] = src[
                            pan_nearest_yi[i], pan_nearest_xi[i]
                        ]
                    else:
                        secondary[iline, i] = _direction(
                            xc[i],
                            yc[iline],
                            xc[pan_nearest_xi[i]],
                            yc[pan_nearest_yi[i]],
                        )
        # Fill skipped regions with nodata value
        for i in range(nx):
            if line_proximity[i] < 0:
                line_proximity[i] = nodata
        proximity[iline] = line_proximity


def _get_coords_for_chunk(xc, yc, coords_block_info, block_info):
    # Trim off band dim if present
    chunk_loc = block_info[0]["chunk-location"][-2:]
    ychunks, xchunks = coords_block_info["chunks"]
    ydepth, xdepth = coords_block_info["depth"]
    xc = da.from_array(xc, chunks=xchunks)
    yc = da.from_array(yc, chunks=ychunks)
    xc = da.overlap.overlap(xc, depth=xdepth, boundary=0)
    yc = da.overlap.overlap(yc, depth=ydepth, boundary=0)
    xc = xc.blocks[chunk_loc[1]].compute()
    yc = yc.blocks[chunk_loc[0]].compute()
    return xc, yc


@single_band_mappable(pass_block_info=True)
def _proximity_analysis_chunk(
    src,
    target_values,
    xc,
    yc,
    coords_block_info,
    max_distance,
    dist_sqr_func,
    full_precision,
    alloc_dtype,
    nodata,
    mode,
    block_info=None,
):
    assert mode in {_MODE_PROXIMITY, _MODE_ALLOCATION, _MODE_DIRECTION}

    # Output arrays
    out_dtype = F64 if full_precision else F32
    prox_dst = np.empty(src.shape, dtype=out_dtype)
    # Only fully allocate array if needed for allocation calculation
    if mode == _MODE_PROXIMITY:
        secondary_dst = np.full((1, 1), nodata, dtype=src.dtype)
    else:
        out_dtype = alloc_dtype if mode == _MODE_ALLOCATION else out_dtype
        secondary_dst = np.full(src.shape, nodata, dtype=out_dtype)

    xc, yc = _get_coords_for_chunk(xc, yc, coords_block_info, block_info)

    _compute_proximity(
        src,
        prox_dst,
        secondary_dst,
        target_values,
        xc,
        yc,
        max_distance,
        dist_sqr_func,
        nodata,
        mode,
    )

    if mode == _MODE_PROXIMITY:
        return prox_dst
    else:
        return secondary_dst


def _validate_lonlat_coords(lon, lat):
    if ((lat < -90) | (lat > 90)).any():
        raise ValueError(
            "Invalid longitude values for great circle distance calculation."
            " Values must be in [-90, 90]."
        )
    if np.any(lon > 360) or np.any(lon < -180):
        raise ValueError(
            "Invalid latitude values for great circle distance calculation."
            " Values must be in [-180, 180] or [0, 360]."
        )
    if np.any(lon > 180) and np.any(lon < 0):
        raise ValueError(
            "Invalid latitude values for great circle distance calculation."
            " Values must be in [-180, 180] or [0, 360]."
        )


def _estimate_min_resolution(lon, lat):
    x1 = lon[0]
    y1 = lat.max()
    x2 = lon[1]
    y2 = y1
    xmin = np.sqrt(_haversine_dist_sqr(x1, y1, x2, y2))
    x1 = lon[0]
    x2 = lon[0]
    # Get two largest latitudes
    y1, y2 = np.partition(lat, -2)[-2:]
    ymin = np.sqrt(_haversine_dist_sqr(x1, y1, x2, y2))
    return xmin, ymin


def _proximity_analysis(
    raster,
    mode,
    target_values,
    distance_metric,
    max_distance,
    double_precision,
):
    raster = get_raster(raster)
    if distance_metric not in _DISTANCE_FUNC_MAP:
        raise ValueError(f"Invalid distance metric: {distance_metric!r}")
    dist_sqr_func = _DISTANCE_FUNC_MAP[distance_metric]
    if target_values is None:
        target_values = []
    target_values = np.atleast_1d(target_values)
    if target_values.ndim > 1:
        raise ValueError("target_values must be 1D")
    if not np.isfinite(target_values).all():
        raise ValueError("target_values must be finite")
    if max_distance <= 0:
        raise ValueError(
            f"max_distance must be greater than 0: {max_distance!r}"
        )

    x = raster.x
    y = raster.y
    if distance_metric in ("haversine", "great_circle"):
        _validate_lonlat_coords(x, y)
    if max_distance >= np.sqrt(dist_sqr_func(x[0], y[0], x[-1], y[-1])):
        # If max distance encompasses the entire raster, rechunk to entire
        # raster.
        _, ny, nx = raster.shape
        raster = Raster(
            raster._ds.chunk({"band": 1, "y": ny, "x": nx}), _fast_path=True
        )
        xdepth = 0
        ydepth = 0
    else:
        if distance_metric not in ("haversine", "great_circle"):
            resolution = raster.resolution
        else:
            # When x/y are lon/lat, the resolution in meters decreases with
            # increasing latitude. Estimate the minimum resolution (cell size
            # at highest latitude) and use that for chunk overlap calculation.
            # Use minimum so that we always overestimate and never
            # underestimate the depth.
            resolution = _estimate_min_resolution(x, y)
        xdepth, ydepth = np.ceil(max_distance / np.abs(resolution)).astype(int)
    coords_block_info = {
        "chunks": raster.data.chunks[1:],
        "depth": (ydepth, xdepth),
    }
    if mode in (_MODE_PROXIMITY, _MODE_DIRECTION):
        out_dtype = F64 if double_precision else F32
    else:
        out_dtype = raster.dtype
    nodata = get_default_null_value(out_dtype)
    orig_raster_dtype = raster.dtype
    # If masked, convert to float and replace null values with nan
    raster = get_raster(raster, null_to_nan=True)

    out_data = da.map_overlap(
        _proximity_analysis_chunk,
        raster.data,
        depth=(0, ydepth, xdepth),
        boundary=np.nan,
        meta=np.array((), dtype=out_dtype),
        # func args
        target_values=target_values,
        xc=x,
        yc=y,
        coords_block_info=coords_block_info,
        max_distance=max_distance,
        dist_sqr_func=dist_sqr_func,
        full_precision=double_precision,
        alloc_dtype=orig_raster_dtype if mode == _MODE_ALLOCATION else None,
        nodata=nodata,
        mode=mode,
    )
    xout = xr.DataArray(
        out_data,
        coords=raster.xdata.coords,
        dims=raster.xdata.dims,
        attrs=raster.xdata.attrs,
    ).rio.write_nodata(nodata)
    rs_out = Raster(xout)
    return rs_out


def pa_proximity(
    raster,
    target_values=None,
    max_distance=np.inf,
    distance_metric="euclidean",
    double_precision=False,
):
    """Compute the proximity of each cell to the closest source cell.

    This function takes a source raster and uses it to compute a proximity
    raster. The proximity raster's cells contain the distances, as determined
    by the `distance_metric` input, to the nearest source cell.
    `distance_metric` selects the function to use for calculating distances
    between cells.

    This is very similar to ESRI's Euclidean Distance tool and GDAL's proximity
    function.

    Parameters
    ----------
    raster : Raster, path str
        The input source raster. If `target_values` is not specified or empty,
        any non-null, finite, non-zero cell is considered a source. If
        `target_values` is specified, it will be used to find source cells with
        the given values.
    target_values : array-like, optional
        A 1D sequence of values to use to find source cells in `raster`. If not
        specified, all non-null, finite, non-zero cells are considered sources.
    max_distance : scalar, optional
        The maximum distance (**in georeferenced units**) to consider when
        computing proximities. Distances greater than this value will not be
        computed and the corresponding cells will be skipped. This value also
        determines how much of `raster` is loaded into memory when carrying out
        the computation. A larger value produces a larger memory footprint. The
        default is infinity which causes the entire raster to be loaded at
        computation time. No cells will be skipped in this case.
    distance_metric : str
        The name of the distance metric to use for computing distances. Default
        is `'euclidean'`.

        'euclidean'
            The euclidean distance between two points: ``sqrt((x1 - x2)^2 + (y1
            - y2)^2)``
        'taxi' 'taxicab' 'manhatten' 'city_block'
            The taxicab distance between two points: ``|x1 - x2| + |y1 - y2|``
        'chebyshev' 'chessboard'
            The chebyshev or chessboard distance: ``max(|x1 - x2|, |y1 - y2|)``
        'haversine' 'great_circle'
            The great circle distance between points on a sphere. This
            implementation uses the WGS84 mean earth radius of 6,371,009 m.
            Coordinates are assumed to be in degrees. Latitude must be in the
            range [-90, 90] and longitude must be in the range [-180, 180].
    double_precision : bool
        If ``True``, distances will will be computed with 64-bit precision,
        otherwise 32-bit floats are used. Default is ``False``.

    Returns
    -------
    Raster
        The proximity raster.

    References
    ----------
    * `ESRI: Proximity Analysis <https://desktop.arcgis.com/en/arcmap/latest/analyze/commonly-used-tools/proximity-analysis.htm>`_
    * `ESRI: Euclidean Distance <https://desktop.arcgis.com/en/arcmap/latest/tools/spatial-analyst-toolbox/euclidean-distance.htm>`_
    * `GDAL: Proximity <https://gdal.org/api/gdal_alg.html#_CPPv420GDALComputeProximity15GDALRasterBandH15GDALRasterBandHPPc16GDALProgressFuncPv>`_
    * `Taxicab Distance <https://en.wikipedia.org/wiki/Taxicab_geometry>`_
    * `Chebyshev Distance <https://en.wikipedia.org/wiki/Chebyshev_distance>`_
    * `Great-Circle Distance <https://en.wikipedia.org/wiki/Great-circle_distance>`_

    See also
    --------
    pa_allocation, pa_direction, proximity_analysis

    """  # noqa: E501
    return _proximity_analysis(
        raster,
        mode=_MODE_PROXIMITY,
        target_values=target_values,
        distance_metric=distance_metric,
        max_distance=max_distance,
        double_precision=double_precision,
    )


def pa_allocation(
    raster,
    target_values=None,
    distance_metric="euclidean",
    max_distance=np.inf,
    double_precision=False,
):
    """
    Compute the proximity allocation of each cell to the closest source cell.

    Allocation assigns each raster cell the value of the nearest source cell.
    This function takes a source raster and uses it to compute a proximity
    allocation raster. For large `max_distance` values, this forms the raster
    analogue of a Voronoi diagram. `distance_metric` selects the function to
    use for calculating distances between cells.

    This is very similar to ESRI's Euclidean Allocation tool.

    Parameters
    ----------
    raster : Raster, path str
        The input source raster. If `target_values` is not specified or empty,
        any non-null, finite, non-zero cell is considered a source. If
        `target_values` is specified, it will be used to find source cells with
        the given values.
    target_values : array-like, optional
        A 1D sequence of values to use to find source cells in `raster`. If not
        specified, all non-null, finite, non-zero cells are considered sources.
    max_distance : scalar, optional
        The maximum distance (**in georeferenced units**) to consider when
        computing proximities. Distances greater than this value will not be
        computed and the corresponding cells will be skipped. This value also
        determines how much of `raster` is loaded into memory when carrying out
        the computation. A larger value produces a larger memory footprint. The
        default is infinity which causes the entire raster to be loaded at
        computation time. No cells will be skipped in this case.
    distance_metric : str
        The name of the distance metric to use for computing distances. Default
        is `'euclidean'`.

        'euclidean'
            The euclidean distance between two points: ``sqrt((x1 - x2)^2 + (y1
            - y2)^2)``
        'taxi' 'taxicab' 'manhatten' 'city_block'
            The taxicab distance between two points: ``|x1 - x2| + |y1 - y2|``
        'chebyshev' 'chessboard'
            The chebyshev or chessboard distance: ``max(|x1 - x2|, |y1 - y2|)``
        'haversine' 'great_circle'
            The great circle distance between points on a sphere. This
            implementation uses the WGS84 mean earth radius of 6,371,009 m.
            Coordinates are assumed to be in degrees. Latitude must be in the
            range [-90, 90] and longitude must be in the range [-180, 180].
    double_precision : bool
        If ``True``, distances will will be computed with 64-bit precision,
        otherwise 32-bit floats are used. Default is ``False``.

    Returns
    -------
    Raster
        The allocation raster.

    References
    ----------
    * `ESRI: Proximity Analysis <https://desktop.arcgis.com/en/arcmap/latest/analyze/commonly-used-tools/proximity-analysis.htm>`_
    * `ESRI: Euclidean Allocation <https://desktop.arcgis.com/en/arcmap/latest/tools/spatial-analyst-toolbox/euclidean-allocation.htm>`_
    * `Voronoi Diagram <https://en.wikipedia.org/wiki/Voronoi_diagram>`_
    * `Taxicab Distance <https://en.wikipedia.org/wiki/Taxicab_geometry>`_
    * `Chebyshev Distance <https://en.wikipedia.org/wiki/Chebyshev_distance>`_
    * `Great-Circle Distance <https://en.wikipedia.org/wiki/Great-circle_distance>`_

    See also
    --------
    pa_proximity, pa_direction, proximity_analysis

    """  # noqa: E501
    return _proximity_analysis(
        raster,
        mode=_MODE_ALLOCATION,
        target_values=target_values,
        distance_metric=distance_metric,
        max_distance=max_distance,
        double_precision=double_precision,
    )


def pa_direction(
    raster,
    target_values=None,
    distance_metric="euclidean",
    max_distance=np.inf,
    double_precision=False,
):
    """
    Compute the direction of each cell to the closest source cell.

    This function takes a source raster and uses it to compute a direction
    raster. Each cell contains the direction in degrees to the nearest source
    cell as determined by the distance metric. `distance_metric` selects the
    function to use for calculating distances between cells. Directions are in
    degrees on the range [0, 360]. 0 indicates that a source cell. 360 is
    North, 90 East, 180 South, and 270 indicates West. They indicate the
    direction **to, not from**, the nearest source cell.

    This is very similar to ESRI's Euclidean Direction tool.

    Parameters
    ----------
    raster : Raster, path str
        The input source raster. If `target_values` is not specified or empty,
        any non-null, finite, non-zero cell is considered a source. If
        `target_values` is specified, it will be used to find source cells with
        the given values.
    target_values : array-like, optional
        A 1D sequence of values to use to find source cells in `raster`. If not
        specified, all non-null, finite, non-zero cells are considered sources.
    max_distance : scalar, optional
        The maximum distance (**in georeferenced units**) to consider when
        computing proximities. Distances greater than this value will not be
        computed and the corresponding cells will be skipped. This value also
        determines how much of `raster` is loaded into memory when carrying out
        the computation. A larger value produces a larger memory footprint. The
        default is infinity which causes the entire raster to be loaded at
        computation time. No cells will be skipped in this case.
    distance_metric : str
        The name of the distance metric to use for computing distances. Default
        is `'euclidean'`.

        'euclidean'
            The euclidean distance between two points: ``sqrt((x1 - x2)^2 + (y1
            - y2)^2)``
        'taxi' 'taxicab' 'manhatten' 'city_block'
            The taxicab distance between two points: ``|x1 - x2| + |y1 - y2|``
        'chebyshev' 'chessboard'
            The chebyshev or chessboard distance: ``max(|x1 - x2|, |y1 - y2|)``
        'haversine' 'great_circle'
            The great circle distance between points on a sphere. This
            implementation uses the WGS84 mean earth radius of 6,371,009 m.
            Coordinates are assumed to be in degrees. Latitude must be in the
            range [-90, 90] and longitude must be in the range [-180, 180].
    double_precision : bool
        If ``True``, distances will will be computed with 64-bit precision,
        otherwise 32-bit floats are used. Default is ``False``.

    Returns
    -------
    Raster
        The direction raster.

    References
    ----------
    * `ESRI: Proximity Analysis <https://desktop.arcgis.com/en/arcmap/latest/analyze/commonly-used-tools/proximity-analysis.htm>`_
    * `ESRI: Euclidean Direction <https://desktop.arcgis.com/en/arcmap/latest/tools/spatial-analyst-toolbox/euclidean-direction.htm>`_
    * `Taxicab Distance <https://en.wikipedia.org/wiki/Taxicab_geometry>`_
    * `Chebyshev Distance <https://en.wikipedia.org/wiki/Chebyshev_distance>`_
    * `Great-Circle Distance <https://en.wikipedia.org/wiki/Great-circle_distance>`_

    See also
    --------
    pa_proximity, pa_allocation, proximity_analysis

    """  # noqa: E501
    return _proximity_analysis(
        raster,
        mode=_MODE_DIRECTION,
        target_values=target_values,
        distance_metric=distance_metric,
        max_distance=max_distance,
        double_precision=double_precision,
    )


def proximity_analysis(
    raster,
    target_values=None,
    distance_metric="euclidean",
    max_distance=np.inf,
    double_precision=False,
):
    """
    Compute the proximity, allocation, and direction for a given source raster.

    This function uses proximity analysis to compute the proximity, allocation,
    and direction rasters for a given source raster. See the individual
    functions for more details:

    * :func:`pa_proximity`
    * :func:`pa_allocation`
    * :func:`pa_direction`

    This is very similar to ESRI's Proximity Analysis tools.

    Parameters
    ----------
    raster : Raster, path str
        The input source raster. If `target_values` is not specified or empty,
        any non-null, finite, non-zero cell is considered a source. If
        `target_values` is specified, it will be used to find source cells with
        the given values.
    target_values : array-like, optional
        A 1D sequence of values to use to find source cells in `raster`. If not
        specified, all non-null, finite, non-zero cells are considered sources.
    max_distance : scalar, optional
        The maximum distance (**in georeferenced units**) to consider when
        computing proximities. Distances greater than this value will not be
        computed and the corresponding cells will be skipped. This value also
        determines how much of `raster` is loaded into memory when carrying out
        the computation. A larger value produces a larger memory footprint. The
        default is infinity which causes the entire raster to be loaded at
        computation time. No cells will be skipped in this case.
    distance_metric : str
        The name of the distance metric to use for computing distances. Default
        is `'euclidean'`.

        'euclidean'
            The euclidean distance between two points: ``sqrt((x1 - x2)^2 + (y1
            - y2)^2)``
        'taxi' 'taxicab' 'manhatten' 'city_block'
            The taxicab distance between two points: ``|x1 - x2| + |y1 - y2|``
        'chebyshev' 'chessboard'
            The chebyshev or chessboard distance: ``max(|x1 - x2|, |y1 - y2|)``
        'haversine' 'great_circle'
            The great circle distance between points on a sphere. This
            implementation uses the WGS84 mean earth radius of 6,371,009 m.
            Coordinates are assumed to be in degrees. Latitude must be in the
            range [-90, 90] and longitude must be in the range [-180, 180].
    double_precision : bool
        If ``True``, distances will will be computed with 64-bit precision,
        otherwise 32-bit floats are used. Default is ``False``.

    Returns
    -------
    Raster
        The direction raster.

    References
    ----------
    * `ESRI: Proximity Analysis <https://desktop.arcgis.com/en/arcmap/latest/analyze/commonly-used-tools/proximity-analysis.htm>`_
    * `ESRI: Euclidean Distance <https://desktop.arcgis.com/en/arcmap/latest/tools/spatial-analyst-toolbox/euclidean-distance.htm>`_
    * `ESRI: Euclidean Allocation <https://desktop.arcgis.com/en/arcmap/latest/tools/spatial-analyst-toolbox/euclidean-allocation.htm>`_
    * `ESRI: Euclidean Direction <https://desktop.arcgis.com/en/arcmap/latest/tools/spatial-analyst-toolbox/euclidean-direction.htm>`_
    * `Taxicab Distance <https://en.wikipedia.org/wiki/Taxicab_geometry>`_
    * `Chebyshev Distance <https://en.wikipedia.org/wiki/Chebyshev_distance>`_
    * `Great-Circle Distance <https://en.wikipedia.org/wiki/Great-circle_distance>`_

    See also
    --------
    pa_proximity, pa_allocation, pa_direction

    """  # noqa: E501
    prox = pa_proximity(
        raster=raster,
        target_values=target_values,
        distance_metric=distance_metric,
        max_distance=max_distance,
        double_precision=double_precision,
    )
    alloc = pa_allocation(
        raster=raster,
        target_values=target_values,
        distance_metric=distance_metric,
        max_distance=max_distance,
        double_precision=double_precision,
    )
    direction = pa_direction(
        raster=raster,
        target_values=target_values,
        distance_metric=distance_metric,
        max_distance=max_distance,
        double_precision=double_precision,
    )
    return prox, alloc, direction



"""
   Description: general module used to perform common spatial analyses
   on Raster objects

   * `ESRI Generalization Tools <https://pro.arcgis.com/en/pro-app/latest/tool-reference/spatial-analyst/an-overview-of-the-generalization-tools.htm>`_
   * `ESRI Local Tools <https://pro.arcgis.com/en/pro-app/latest/tool-reference/spatial-analyst/an-overview-of-the-local-tools.htm>`_

"""  # noqa: E501
import os
import re
from collections.abc import Iterable, Sequence
from functools import partial

import dask.array as da
import numba as nb
import numpy as np
import xarray as xr
from dask_image import ndmeasure as ndm
from scipy.ndimage import (
    binary_dilation,
    binary_erosion,
    grey_dilation,
    grey_erosion,
)

from raster_tools.creation import empty_like, zeros_like
from raster_tools.dtypes import (
    BOOL,
    F16,
    F32,
    F64,
    I8,
    I16,
    I32,
    I64,
    U8,
    U16,
    U32,
    U64,
    get_common_dtype,
    is_bool,
    is_int,
    is_scalar,
    is_str,
)
from raster_tools.focal import _get_offsets
from raster_tools.masking import (
    create_null_mask,
    get_default_null_value,
    reconcile_nullvalue_with_dtype,
)
from raster_tools.raster import Raster, get_raster
from raster_tools.stat_common import (
    nan_unique_count_jit,
    nanargmax_jit,
    nanargmin_jit,
    nanasm_jit,
    nanentropy_jit,
    nanmode_jit,
)
from raster_tools.utils import make_raster_ds

__all__ = [
    "aggregate",
    "band_concat",
    "dilate",
    "erode",
    "local_stats",
    "predict_model",
    "reclassify",
    "regions",
    "remap_range",
]

# TODO: mosaic


def _create_labels(xarr, wd, uarr=None):
    if uarr is None:
        uarr = da.unique(xarr).compute()
    # Drop 0 so they are skipped later
    uarr = uarr[uarr != 0]

    cum_num = 0
    result = da.zeros_like(xarr)
    for v in uarr:
        labeled_array, num_features = ndm.label((xarr == v), structure=wd)
        result += da.where(
            labeled_array > 0, labeled_array + cum_num, 0
        ).astype(result.dtype)
        cum_num += num_features
    return result


def regions(raster, neighbors=4, unique_values=None):
    """Calculates the unique regions (patches) within a raster band.

    The approach is based on ESRI's region group calculation.

    Parameters
    ----------
    raster : Raster or path str
        The raster to perform the calculation on. All unique non zero values
        will be used to define unique regions.
    neighbors : int, optional
        The neighborhood connectivity value. Valid values are 4 and 8. If
        4, a rook pattern is used, e.g. the neighbors along the horizontal and
        vertical directions are used. if 8, then all of the 8 neighbors are
        used. Default is 4.
    unique_values : array or list, optional
        Values that represent zones from which regions will be made. Values not
        included will be grouped together to form one large zone. If `None`,
        each unique value in the raster will be considered a zone and will be
        calculated up front.

    Returns
    -------
    Raster
        The resulting raster of unique regions values. The bands will have the
        same shape as the original Raster. The null value mask from the origial
        raster will be applied to the result.

    References
    ----------
    * `ESRI: Region Group <https://pro.arcgis.com/en/pro-app/latest/tool-reference/spatial-analyst/region-group.htm>`_

    """  # noqa: E501
    raster = get_raster(raster)
    rs_out = zeros_like(raster, dtype=U64)

    if not is_int(neighbors):
        raise TypeError(
            f"neighbors argument must be an int. Got {type(neighbors)}"
        )
    if neighbors == 4:
        wd = np.array([[0, 1, 0], [1, 1, 1], [0, 1, 0]])
    elif neighbors == 8:
        wd = np.ones((3, 3), dtype=int)
    else:
        raise ValueError(
            f"Invalid value for neighbors parameter: {repr(neighbors)}"
        )

    if unique_values is not None:
        if isinstance(unique_values, (np.ndarray, da.Array, Sequence)):
            unique_values = np.asarray(unique_values)
        else:
            raise TypeError("Invalid type for unique_values parameter")

    data = raster.data
    dout = rs_out.data
    if raster._masked:
        # Set null values to 0 to skip them in the labeling phase
        data = da.where(raster._ds.mask.data, 0, data)

    for bnd in range(data.shape[0]):
        dout[bnd] = _create_labels(data[bnd], wd, unique_values)

    if raster._masked:
        nv = reconcile_nullvalue_with_dtype(raster.null_value, rs_out.dtype)
        dout = da.where(raster._ds.mask.data, nv, dout)
    rs_out._ds.raster.data = dout
    return rs_out


@nb.jit(nopython=True, nogil=True, parallel=True)
def _coarsen_chunk(x, axis, func, out_dtype, check_nan):
    dims = sorted(set(range(len(x.shape))) - set(axis))
    shape = (x.shape[dims[0]], x.shape[dims[1]], x.shape[dims[2]])
    out = np.empty(shape, out_dtype)
    for i in range(shape[0]):
        for j in nb.prange(shape[1]):
            for k in nb.prange(shape[2]):
                v = func(x[i, j, :, k, :])
                if check_nan and np.isnan(v):
                    # It doesn't matter what value is swapped with nan values
                    # since those cells will be masked out later.
                    v = 0
                out[i, j, k] = v
    return out


def _coarsen_block_map(x, axis, agg_func, out_dtype, check_nan):
    dims = sorted(set(range(len(x.shape))) - set(axis))
    chunks = tuple(x.chunks[d] for d in dims)
    return da.map_blocks(
        partial(
            _coarsen_chunk,
            axis=axis,
            func=agg_func,
            out_dtype=out_dtype,
            check_nan=check_nan,
        ),
        x,
        chunks=chunks,
        drop_axis=axis,
        meta=np.array((), dtype=out_dtype),
    )


def _get_coarsen_dtype(stat, window_size, input_dtype):
    if stat == "unique":
        return np.min_scalar_type(window_size)
    if stat in ("mode", "min", "max"):
        return input_dtype
    if input_dtype == F32:
        return F32
    return F64


_COARSEN_STYPE_TO_FUNC = {
    "max": lambda x: x.max(),
    "mean": lambda x: x.mean(),
    "median": lambda x: x.median(),
    "min": lambda x: x.min(),
    "prod": lambda x: x.prod(),
    "std": lambda x: x.std(),
    "sum": lambda x: x.sum(),
    "var": lambda x: x.var(),
}
_COARSEN_STYPE_TO_CUSTOM_FUNC = {
    "asm": nanasm_jit,
    "entropy": nanentropy_jit,
    "mode": nanmode_jit,
    "unique": nan_unique_count_jit,
}


def _get_unique_dtype(cur_dtype):
    if cur_dtype in (I8, U8):
        return I16
    if cur_dtype == U16:
        return I32
    if cur_dtype == U32:
        return I64
    return cur_dtype


def aggregate(raster, expand_cells, stype):
    """Creates a Raster of aggregated cell values for a new resolution.

    The approach is based on ESRI's aggregate and majority filter functions.

    Parameters
    ----------
    raster : Raster or path str
        Input Raster object or path string
    expand_cells : 2-tuple, list, array-like
        Tuple, array, or list of the number of cells to expand in y and x
        directions. The first element corresponds to the y dimension and the
        second to the x dimension.
    stype : str
        Summarization type. Valid opition are mean, std, var, max, min,
        unique prod, median, mode, sum, unique, entropy, asm.

    Returns
    -------
    Raster
        The resulting raster of aggregated values.

    References
    ----------
    * `ESRI aggregate <https://pro.arcgis.com/en/pro-app/latest/tool-reference/spatial-analyst/aggregate.htm>`_
    * `ESRI majority filter <https://pro.arcgis.com/en/pro-app/latest/tool-reference/spatial-analyst/majority-filter.htm>`_

    """  # noqa: E501
    expand_cells = np.atleast_1d(expand_cells)
    if not is_int(expand_cells.dtype):
        raise TypeError("expand_cells must contain integers")
    if expand_cells.shape != (2,):
        raise ValueError("expand_cells must contain 2 elements")
    if (expand_cells == 1).all():
        raise ValueError("expand_cells values cannont both be one")
    if not (expand_cells >= 1).all():
        raise ValueError("All expand_cells values must be >= 1")
    if not is_str(stype):
        raise TypeError("stype argument must be a string")
    stype = stype.lower()
    if (
        stype not in _COARSEN_STYPE_TO_FUNC
        and stype not in _COARSEN_STYPE_TO_CUSTOM_FUNC
    ):
        raise ValueError(f"Invalid stype argument: {repr(stype)}")
    orig_dtype = get_raster(raster).dtype
    rs = get_raster(raster, null_to_nan=True)

    xda = rs.xdata
    xmask = rs.xmask
    dim_map = {"y": expand_cells[0], "x": expand_cells[1]}
    xdac = xda.coarsen(dim=dim_map, boundary="trim")
    if stype in _COARSEN_STYPE_TO_FUNC:
        xda = _COARSEN_STYPE_TO_FUNC[stype](xdac)
    else:
        custom_stat_func = _COARSEN_STYPE_TO_CUSTOM_FUNC[stype]
        check_nan = stype == "mode"
        out_dtype = _get_coarsen_dtype(
            stype, np.prod(expand_cells), orig_dtype
        )
        # Use partial because reduce seems to be bugged and grabs kwargs that
        # it shouldn't.
        xda = xdac.reduce(
            partial(
                _coarsen_block_map,
                agg_func=custom_stat_func,
                out_dtype=out_dtype,
                check_nan=check_nan,
            )
        )
    # Coarsen mask as well
    if rs._masked:
        xmask = rs.xmask.coarsen(dim=dim_map, boundary="trim").all()

    ds_out = make_raster_ds(xda, xmask)
    if rs._masked:
        if stype == "unique":
            ds_out["raster"] = ds_out.raster.astype(
                _get_unique_dtype(ds_out.raster.dtype)
            )
        elif stype == "mode":
            # Cast back to original dtype. Original null value will work
            ds_out["raster"] = ds_out.raster.astype(orig_dtype)
        # Replace null cells with null value acording to mask
        nv = get_default_null_value(ds_out.raster.dtype)
        ds_out["raster"] = xr.where(xmask, nv, ds_out.raster).rio.write_nodata(
            nv
        )
    else:
        ds_out["mask"] = xr.zeros_like(rs.get_bands(1).xmask)
    if rs.crs is not None:
        ds_out = ds_out.rio.write_crs(rs.crs)
    return Raster(ds_out, _fast_path=True)


def predict_model(raster, model):
    """
    Predict cell estimates from a model using raster band values as predictors.

    Predictor bands correspond to the order of the predictor variables within
    the model. Outputs are raster surfaces with bands cell values depending on
    the type of model.

    The function uses the `model` class' predict method to estimate a new
    raster surface.

    Parameters
    ----------
    raster : Raster or path str
        Raster of predictor variables where the bands correspond to variables
        in the model (one band for each variable in the model).
    model : object
        The model used to estimate new values. Must have a `predict` method
        that takes an `xarray.DataArray` object. The provided DataArray will
        have null cells marked with ``NaN``.

    Returns
    -------
    Raster
        The resulting raster of estmated values.

    """
    rs = get_raster(raster, null_to_nan=True)
    xarr = rs.xdata
    xarrout = model.predict(xarr)

    return Raster(xarrout)


@nb.jit(nopython=True, nogil=True)
def _local_chunk(x, func, out):
    dshape = x.shape
    rws = dshape[1]
    clms = dshape[2]
    for r in range(rws):
        for c in range(clms):
            farr = x[:, r, c]
            out[0, r, c] = func(farr)
    return out


def _get_local_chunk_func(func, out_dtype):
    def wrapped(x):
        return _local_chunk(
            x, func, np.empty((1, *x.shape[1:]), dtype=out_dtype)
        )

    return wrapped


_LOCAL_STYPE_TO_NUMPY_FUNC = {
    "mean": np.nanmean,
    "std": np.nanstd,
    "var": np.nanvar,
    "max": np.nanmax,
    "min": np.nanmin,
    "prod": np.nanprod,
    "sum": np.nansum,
    "median": np.nanmedian,
}
# Use custom min/max band because numpy versions throw errors if all values are
# nan.
_LOCAL_STYPE_TO_CUSTOM_FUNC = {
    "asm": nanasm_jit,
    "entropy": nanentropy_jit,
    "minband": nanargmin_jit,
    "maxband": nanargmax_jit,
    "mode": nanmode_jit,
    "unique": nan_unique_count_jit,
}


def _local(data, stype):
    bnds = data.shape[0]
    orig_chunks = data.chunks
    # Rechunk so band dim is contiguous and chunk sizes are reasonable
    data = da.rechunk(data, chunks=(bnds, "auto", "auto"))

    ffun = _LOCAL_STYPE_TO_CUSTOM_FUNC[stype]
    if stype == "unique":
        out_dtype = np.min_scalar_type(data.shape[0])
    elif stype == "mode":
        out_dtype = data.dtype
    elif stype in ("minband", "maxband"):
        # min/max band inner funcs will return < 0 if all values are nan which
        # will cause overflow for the unsigned type returned by
        # min_scalar_type. This is fine since we can mask them out later.
        out_dtype = np.min_scalar_type(data.shape[0] - 1)
    elif data.dtype == F32:
        out_dtype = F32
    else:
        out_dtype = F64
    ffun = _get_local_chunk_func(ffun, out_dtype)

    out_chunks = ((1,), *data.chunks[1:])
    data_out = da.map_blocks(
        ffun,
        data,
        chunks=out_chunks,
        dtype=out_dtype,
        meta=np.array((), dtype=out_dtype),
    )
    # Rechunk again to expand/contract chunks to a reasonable size and split
    # bands apart
    return da.rechunk(data_out, chunks=(1, *orig_chunks[1:]))


def local_stats(raster, stype):
    """Creates a Raster of summarized values across bands.

    The approach is based on ESRI's local function.

    Parameters
    ----------
    raster : Raster or path str
        Input Raster object or path string
    stype : str
        Summarization type. Valid opition are mean, std, var, max, min,
        maxband, minband, prod, sum, mode, median, unique, entropy, asm

    Returns
    -------
    Raster
        The resulting raster of values aggregated along the band dimension.

    References
    ----------
    * `ESRI local <https://pro.arcgis.com/en/pro-app/latest/tool-reference/spatial-analyst/an-overview-of-the-local-tools.htm>`_

    """  # noqa: E501
    if not is_str(stype):
        raise TypeError("stype argument must be a string")
    stype = stype.lower()
    if (
        stype not in _LOCAL_STYPE_TO_NUMPY_FUNC
        and stype not in _LOCAL_STYPE_TO_CUSTOM_FUNC
    ):
        raise ValueError(f"Invalid stype aregument: {repr(stype)}")
    orig_dtype = get_raster(raster).dtype
    rs = get_raster(raster, null_to_nan=True)

    xda = rs.xdata.copy()
    xmask = rs.xmask
    if stype in _LOCAL_STYPE_TO_NUMPY_FUNC:
        xndata = xda.reduce(
            _LOCAL_STYPE_TO_NUMPY_FUNC[stype], dim="band", keepdims=True
        )
    else:
        data = xda.data
        xndata = empty_like(rs.get_bands(1)).xdata
        xndata.data = _local(data, stype)
    if rs._masked:
        xmask = xmask.reduce(
            np.all,
            dim="band",
            keepdims=True,
        )
    else:
        xmask = xr.zeros_like(xmask, dtype=bool)

    ds_out = make_raster_ds(xndata, xmask)
    if rs._masked:
        if stype in ("unique", "minband", "maxband"):
            ds_out["raster"] = ds_out.raster.astype(
                _get_unique_dtype(ds_out.raster.dtype)
            )
        elif stype == "mode":
            # Cast back to original dtype. Original null value will work
            ds_out["raster"] = ds_out.raster.astype(orig_dtype)
        nv = get_default_null_value(ds_out.raster.dtype)
        ds_out["raster"] = ds_out.raster.rio.write_nodata(nv)
        ds_out.raster.data = da.where(ds_out.mask.data, nv, ds_out.raster.data)
    if rs.crs is not None:
        ds_out = ds_out.rio.write_crs(rs.crs)
    return Raster(ds_out, _fast_path=True)


def _morph_op_chunk(x, footprint, cval, morph_op, binary=False):
    if x.ndim > 2:
        x = x[0]
    if not binary:
        if morph_op == "dilation":
            morph_func = grey_dilation
        else:
            morph_func = grey_erosion
        out = morph_func(x, footprint=footprint, mode="constant", cval=cval)
    else:
        if morph_op == "dilation":
            morph_func = binary_dilation
        else:
            morph_func = binary_erosion
        out = morph_func(x, structure=footprint)
    return out[None]


def _get_footprint(size):
    if isinstance(size, Iterable):
        size = tuple(size)
        if len(size) != 2:
            raise ValueError("size sequence must have lenght 2.")
        if not all(is_int(s) for s in size):
            raise TypeError("size sequence must only contain ints.")
    elif is_int(size):
        size = (size, size)
    else:
        raise TypeError("size input must be an int or sequence of ints.")
    if not all(s > 0 for s in size):
        raise ValueError("size values must be greater than 0.")
    if all(s == 1 for s in size):
        raise ValueError("At least one size value must be greater than 1.")
    footprint = np.ones(size) > 0
    return footprint


def _get_fill(dtype, op):
    if is_int(dtype):
        type_info = np.iinfo(dtype)
    else:
        type_info = np.finfo(dtype)
    if op == "erosion":
        fill = type_info.max
    else:
        # dilation
        fill = type_info.min
    return fill


def _erosion_or_dilation_filter(rs, footprint, op):
    data = rs.data
    fill = _get_fill(rs.dtype, op)
    if rs._masked:
        data = da.where(rs.mask, fill, data)
    rpad, cpad = _get_offsets(footprint.shape)
    # Take max because map_overlap does not support asymmetrical overlaps when
    # a boundary value is given
    depth = {0: 0, 1: max(rpad), 2: max(cpad)}
    data = da.map_overlap(
        partial(
            _morph_op_chunk,
            footprint=footprint,
            cval=fill,
            morph_op=op,
        ),
        data,
        depth=depth,
        boundary=fill,
        dtype=rs.dtype,
        meta=np.array((), dtype=rs.dtype),
    )
    mask = rs.mask
    if rs._masked:
        mask = da.map_overlap(
            partial(
                _morph_op_chunk,
                footprint=footprint,
                cval=fill,
                morph_op=op,
                binary=True,
            ),
            ~mask,
            depth=depth,
            boundary=False,
            dtype=BOOL,
            meta=np.array((), dtype=BOOL),
        )
        mask = ~mask
        data = da.where(mask, rs.null_value, data)
    else:
        mask = mask.copy()
    xrs_out = xr.zeros_like(rs.xdata).rio.write_nodata(rs.null_value)
    xrs_out.data = data
    xmask = xr.DataArray(mask, coords=xrs_out.coords, dims=xrs_out.dims)
    ds_out = make_raster_ds(xrs_out, xmask)
    if rs.crs is not None:
        ds_out = ds_out.rio.write_crs(rs.crs)
    return Raster(ds_out, _fast_path=True)


def dilate(raster, size):
    """Perform dilation on a raster

    Dilation increases the thickness of raster features. Features with higher
    values will cover features with lower values. At each cell, the miximum
    value within a window, defined by `size`, is stored in the output location.
    This is very similar to the max focal filter except that raster features
    are dilated (expanded) into null regions. Dilation is performed on each
    band separately.

    Parameters
    ----------
    raster : Raster or path str
        The raster to dilate
    size : int or 2-tuple of ints
        The shape of the window to use when dilating. A Single int is
        interpreted as the size of a square window. A tuple of 2 ints is used
        as the dimensions of rectangular window. At least one value must be
        greater than 1. Values cannot be less than 1.

    Returns
    -------
    Raster
        The resulting raster with eroded features. This raster will have the
        same shape and meta data as the original

    See also
    --------
    erode, raster_tools.focal.focal

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Dilation_%28morphology%29
    .. [2] https://en.wikipedia.org/wiki/Mathematical_morphology

    """
    raster = get_raster(raster)
    footprint = _get_footprint(size)

    return _erosion_or_dilation_filter(raster, footprint, "dilation")


def erode(raster, size):
    """Perform erosion on a raster

    Erosion increases the thickness of raster features. Features with higher
    values will cover features with lower values. At each cell, the miximum
    value within a window, defined by `size`, is stored in the output location.
    This is very similar to the max focal filter except that raster features
    are eroded (contracted) into null regions. Erosion is performed on each
    band separately.

    Parameters
    ----------
    raster : Raster or path str
        The raster to erode
    size : int or 2-tuple of ints
        The shape of the window to use when eroding. A Single int is
        interpreted as the size of a square window. A tuple of 2 ints is used
        as the dimensions of rectangular window. At least one value must be
        greater than 1. Values cannot be less than 1.

    Returns
    -------
    Raster
        The resulting raster with eroded features. This raster will have the
        same shape and meta data as the original

    See also
    --------
    dilate, raster_tools.focal.focal

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Erosion_%28morphology%29
    .. [2] https://en.wikipedia.org/wiki/Mathematical_morphology

    """
    raster = get_raster(raster)
    footprint = _get_footprint(size)

    return _erosion_or_dilation_filter(raster, footprint, "erosion")


def band_concat(rasters):
    """Join a sequence of rasters along the band dimension.

    Parameters
    ----------
    rasters : sequence of Rasters and/or paths
        The rasters to concatenate. These can be a mix of Rasters and paths.
        All rasters must have the same shape in the x and y dimensions.

    Returns
    -------
    Raster
        The resulting concatenated Raster.

    """
    rasters = [get_raster(raster) for raster in rasters]
    if not rasters:
        raise ValueError("No rasters provided")
    if len(rasters) == 1:
        return rasters[0]

    # TODO: make join a user option or set join="exact"?
    ds = xr.concat([r._ds for r in rasters], dim="band", join="inner")
    ds["band"] = np.arange(np.sum([r.nbands for r in rasters])) + 1
    if any(r._masked for r in rasters):
        nv = get_default_null_value(ds.raster.dtype)
        ds["raster"] = xr.where(ds.mask, nv, ds.raster).rio.write_nodata(nv)
    crs = ([None] + [r.crs for r in rasters if r.crs is not None]).pop()
    if ds.rio.crs is None and crs is not None:
        ds = ds.rio.write_crs(crs)

    return Raster(ds, _fast_path=True)


@nb.jit(nopython=True, nogil=True)
def _remap_values(x, mask, mappings, inclusivity):
    outx = np.zeros_like(x)
    bands, rows, columns = x.shape
    rngs = mappings.shape[0]
    for bnd in range(bands):
        for rw in range(rows):
            for cl in range(columns):
                if mask[bnd, rw, cl]:
                    continue
                vl = int(x[bnd, rw, cl])
                remap = False
                for imap in range(rngs):
                    left, right, new = mappings[imap]
                    if inclusivity == 0:
                        remap = left <= vl < right
                    elif inclusivity == 1:
                        remap = left < vl <= right
                    elif inclusivity == 2:
                        remap = left <= vl <= right
                    elif inclusivity == 3:
                        remap = left < vl < right
                    if remap:
                        outx[bnd, rw, cl] = new
                        break
                if not remap:
                    outx[bnd, rw, cl] = x[bnd, rw, cl]
    return outx


def _normalize_mappings(mappings):
    if not isinstance(mappings, (list, tuple)):
        raise TypeError(
            "Mappings must be either single 3-tuple or list of 3-tuples of "
            "scalars"
        )
    if not len(mappings):
        raise ValueError("No mappings provided")
    if len(mappings) and is_scalar(mappings[0]):
        mappings = [mappings]
    try:
        mappings = [list(m) for m in mappings]
    except TypeError:
        raise TypeError(
            "Mappings must be either single 3-tuple or list of 3-tuples of "
            "scalars"
        )
    for m in mappings:
        if len(m) != 3:
            raise ValueError(
                "Mappings must be either single 3-tuple or list of 3-tuples of"
                " scalars"
            )
        if not all(is_scalar(mi) for mi in m):
            raise TypeError("Mappings values must be scalars")
        if any(np.isnan(mi) for mi in m[:2]):
            raise ValueError("Mapping min and max values cannot be NaN")
        if m[0] >= m[1]:
            raise ValueError(
                "Mapping min value must be strictly less than max value:"
                f" {m[0]}, {m[1]}"
            )
    return mappings


def remap_range(raster, mapping, inclusivity="left"):
    """Remaps values based on a mapping or list of mappings.

    Mappings are applied all at once with earlier mappings taking
    precedence.

    Parameters
    ----------
    raster : Raster or str
        Path string or Raster to perform remap on.
    mapping : 3-tuple of scalars or list of 3-tuples of scalars
        A tuple or list of tuples containing ``(min, max, new_value)``
        scalars. The mappiing(s) map values between the min and max to the
        ``new_value``. If `mapping` is a list and there are mappings that
        conflict or overlap, earlier mappings take precedence. `inclusivity`
        determines which sides of the range are inclusive and exclusive.
    inclusivity : str, optional
        Determines whether to be inclusive or exclusive on either end of the
        range. Default is `'left'`.

        'left' [min, max)
            Left (min) side is inclusive and right (max) side is exclusive.
        'right' (min, max]
            Left (min) side is exclusive and right (max) side is inclusive.
        'both' [min, max]
            Both sides are inclusive.
        'none' (min, max)
            Both sides are exclusive.

    Returns
    -------
    Raster
        The resulting Raster.

    """
    raster = get_raster(raster)
    mappings = _normalize_mappings(mapping)
    if not is_str(inclusivity):
        raise TypeError(
            f"inclusivity must be a str. Got type: {type(inclusivity)}"
        )
    inc_map = {
        name: value
        for name, value in zip(("left", "right", "both", "none"), range(4))
    }
    if inclusivity not in inc_map:
        raise ValueError(f"Invalid inclusivity value. Got: {inclusivity!r}")
    mappings_common_dtype = get_common_dtype([m[-1] for m in mappings])
    out_dtype = np.promote_types(raster.dtype, mappings_common_dtype)
    # numba doesn't understand f16 so use f32 and then downcast
    f16_workaround = out_dtype == F16
    mappings = np.atleast_2d(mappings)

    outrs = raster.copy()
    if out_dtype != outrs.dtype:
        if not f16_workaround:
            outrs = outrs.astype(out_dtype)
        else:
            outrs = outrs.astype(F32)
    elif f16_workaround:
        outrs = outrs.astype(F32)
    data = outrs.data
    func = partial(
        _remap_values, mappings=mappings, inclusivity=inc_map[inclusivity]
    )
    outrs.xdata.data = data.map_blocks(
        func,
        raster.mask,
        dtype=data.dtype,
        meta=np.array((), dtype=data.dtype),
    )
    if f16_workaround:
        outrs = outrs.astype(F16)
    return outrs


def where(condition, true_rast, false_rast):
    """
    Return elements chosen from `true_rast` or `false_rast` depending on
    `condition`.

    Parameters
    ----------
    condition : str or Raster
        A boolean or int raster that indicates where elements in the result
        should be selected from. If the condition is an int raster, it is
        coerced to bool using `condition > 0`.  ``True`` cells pull values from
        `true_rast` and ``False`` cells pull from `y`. *str* is treated as a
        path to a raster.
    true_rast : scalar, Raster, str
        Raster or scalar to pull from if the corresponding location in
        `condition` is ``True``.
    false_rast : scalar, Raster, str
        Raster or scalar to pull from if the corresponding location in
        `condition` is ``False``.

    Returns
    -------
    Raster
        The resulting Raster.

    """
    condition = get_raster(condition)
    if not is_bool(condition.dtype) and not is_int(condition.dtype):
        raise TypeError(
            "Condition argument must be a boolean or integer raster"
        )
    args = []
    for r, name in [(true_rast, "true_rast"), (false_rast, "false_rast")]:
        if not is_scalar(r):
            try:
                r = get_raster(r)
            except TypeError:
                raise TypeError(
                    f"Could not understand {name} argument. Got: {r!r}"
                )
        args.append(r)
    true_rast, false_rast = args
    out_crs = None
    for r in [condition, true_rast, false_rast]:
        crs = getattr(r, "crs", None)
        if crs is not None:
            out_crs = crs
            break

    xtrue, xfalse = [r.xdata if isinstance(r, Raster) else r for r in args]
    masked = any(r._masked if isinstance(r, Raster) else False for r in args)
    scalar_and_nan = all(is_scalar(r) for r in args) and np.isnan(args).any()
    masked |= scalar_and_nan
    xcondition = condition.xdata
    if is_int(condition.dtype):
        # if condition.dtype is not bool then must be an int raster so
        # assume that condition is raster of 0 and 1 values.
        # condition > 0 will grab all 1/True values.
        xcondition = xcondition > 0

    out_xrs = xr.where(xcondition, xtrue, xfalse)
    if masked and not scalar_and_nan:
        xtrue_mask, xfalse_mask = [
            r.xmask
            if isinstance(r, Raster)
            else xr.DataArray(
                create_null_mask(condition.xdata, None),
                dims=condition.xdata.dims,
                coords=condition.xdata.coords,
            )
            for r in args
        ]
        xmask = xr.where(xcondition, xtrue_mask, xfalse_mask)
    elif scalar_and_nan:
        xmask = np.isnan(out_xrs)
    else:
        xmask = xr.DataArray(
            create_null_mask(condition.xdata, None),
            dims=condition.xdata.dims,
            coords=condition.xdata.coords,
        )
    if masked or scalar_and_nan:
        nv = get_default_null_value(out_xrs.dtype)
        out_xrs = xr.where(xmask, nv, out_xrs).rio.write_nodata(nv)
    out_ds = make_raster_ds(out_xrs, xmask)
    if out_crs is not None:
        out_ds = out_ds.rio.write_crs(out_crs)
    return Raster(out_ds, _fast_path=True)


@nb.jit(nopython=True, nogil=True)
def _reclassify_chunk(x, mask, mapping_array, unmapped_to_null, null):
    mapping = dict()
    for i in range(mapping_array.shape[0]):
        mapping[mapping_array[i, 0]] = mapping_array[i, 1]

    out = np.empty_like(x)
    nb, ny, nx = x.shape
    for b in range(nb):
        for i in range(ny):
            for j in range(nx):
                if mask[b, i, j]:
                    out[b, i, j] = null
                    break
                v = x[b, i, j]
                reclass = v in mapping
                if reclass:
                    out[b, i, j] = mapping[v]
                elif unmapped_to_null:
                    out[b, i, j] = null
                else:
                    out[b, i, j] = v
    return out


class RemapFileParseError(Exception):
    pass


_REMAPPING_LINE_PATTERN = re.compile(r"^(?P<from>[+-]*\d+):(?P<to>[+-]*\d+)$")


def _parse_ascii_remap_file(path):
    mapping = {}
    with open(path) as fd:
        for line in fd:
            line = line.strip()
            m = _REMAPPING_LINE_PATTERN.match(line)
            if m is None:
                raise RemapFileParseError(f"Invalid remap line: {line!r}")
            k = int(m.group("from"))
            v = int(m.group("to"))
            if k in mapping:
                raise ValueError("Found duplicate mapping: '{k}:{v}'.")
            mapping[k] = v
    return mapping


def _get_remapping(mapping):
    if is_str(mapping):
        if os.path.exists(mapping):
            mapping = _parse_ascii_remap_file(mapping)
        else:
            raise IOError(f"No such file: {mapping!r}")
    elif not isinstance(mapping, dict):
        raise TypeError(
            f"Remapping must be a str or dict. Got: {type(mapping)!r}"
        )
    if not all(is_int(k) for k in mapping.keys()) or not all(
        is_int(v) for v in mapping.values()
    ):
        raise TypeError("Remapping values must all be integer types")
    return mapping


def reclassify(raster, remapping, unmapped_to_null=False):
    """Reclassify the input raster values based on a mapping.

    This function only works with integer type rasters.

    Parameters
    ----------
    raster : str, Raster
        The input raster to reclassify. Can be a path string or Raster object.
        The raster dtype must be integer.
    remapping : str, dict
        Can be either a ``dict`` or a path string. If a ``dict`` is provided,
        the keys will be reclassified to the corresponding values. If a path
        string, it is treated as an ASCII remap file where each line looks like
        ``a:b`` and ``a`` and ``b`` are integers. All remap values (both from
        and to) must be integers.
    unmapped_to_null : bool, optional
        If ``True``, values not included in the mapping are instead mapped to
        the null value. Default is ``False``.

    Returns
    -------
    Raster
        The remapped raster.

    """
    raster = get_raster(raster)
    remapping = _get_remapping(remapping)
    if not is_int(raster.dtype):
        raise TypeError(
            f"Input raster must be an integer type. Got {raster.dtype!r}"
        )

    out_dtype = raster.dtype
    to_values = list(remapping.values())
    if any(not np.can_cast(v, out_dtype) for v in to_values):
        extra_min = raster.dtype.type(np.iinfo(raster.dtype).min)
        extra_max = raster.dtype.type(np.iinfo(raster.dtype).max)
        out_dtype = get_common_dtype(to_values + [extra_min, extra_max, -1])
    if unmapped_to_null:
        if raster._masked:
            nv = raster.null_value
        else:
            nv = get_default_null_value(out_dtype)
    else:
        nv = 0

    mapping = np.array(list(remapping.items()))
    data = da.map_blocks(
        _reclassify_chunk,
        raster.data.copy().astype(out_dtype),
        raster.mask,
        mapping_array=mapping,
        unmapped_to_null=unmapped_to_null,
        null=nv,
        dtype=out_dtype,
        meta=np.array((), dtype=out_dtype),
    )
    xdata = xr.DataArray(
        data, coords=raster.xdata.coords, dims=raster.xdata.dims
    ).rio.write_nodata(nv)
    if raster.crs is not None:
        xdata = xdata.rio.write_crs(raster.crs)
    return Raster(xdata)
